""" pytest for WorkflowTasks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name

from mock import MagicMock
import pytest

from DIRAC import gLogger
from DIRAC.Interfaces.API.Job import Job

# sut
from DIRAC.TransformationSystem.Client.WorkflowTasks import WorkflowTasks

gLogger.setLevel("DEBUG")

mockTransClient = MagicMock()
mockTransClient.setTaskStatusAndWmsID.return_value = {"OK": True}

WMSClientMock = MagicMock()
jobMonitoringClient = MagicMock()

wfTasks = WorkflowTasks(
    transClient=mockTransClient,
    submissionClient=WMSClientMock,
    jobMonitoringClient=jobMonitoringClient,
    outputDataModule="mock",
)
odm_o = MagicMock()
odm_o.execute.return_value = {"OK": True, "Value": {}}
wfTasks.outputDataModule_o = odm_o

taskDict = {
    1: {"TransformationID": 1, "a1": "aa1", "b1": "bb1", "Site": "MySite"},
    2: {"TransformationID": 1, "a2": "aa2", "b2": "bb2", "InputData": ["a1", "a2"]},
    3: {"TransformationID": 2, "a3": "aa3", "b3": "bb3"},
}

taskDictSimple = {1: {"TransformationID": 1, "Site": "MySite"}, 2: {"TransformationID": 1}, 3: {"TransformationID": 2}}

taskDictSimpleInputs = {1: {"TransformationID": 1, "InputData": ["a1", "a2", "a3"]}}

taskDictNoInputs = {
    1: {"TransformationID": 1, "a1": "aa1", "b1": "bb1", "Site": "MySite"},
    2: {"TransformationID": 1, "a2": "aa2", "b2": "bb2"},
    3: {"TransformationID": 2, "a3": "aa3", "b3": "bb3"},
}

taskDictNoInputsNoSite = {1: {"TransformationID": 1}, 2: {"TransformationID": 1}, 3: {"TransformationID": 2}}

expected = {
    "OK": True,
    "Value": {
        1: {"a1": "aa1", "TaskObject": "", "TransformationID": 1, "b1": "bb1", "Site": "MySite", "JobType": "User"},
        2: {
            "TaskObject": "",
            "a2": "aa2",
            "TransformationID": 1,
            "InputData": ["a1", "a2"],
            "b2": "bb2",
            "Site": "MySite",
            "JobType": "User",
        },
        3: {"TaskObject": "", "a3": "aa3", "TransformationID": 2, "b3": "bb3", "Site": "MySite", "JobType": "User"},
    },
}

expectedBulkSimple = {
    "OK": True,
    "Value": {"BulkJobObject": "", 1: {"TransformationID": 1, "InputData": ["a1", "a2", "a3"], "JobType": "User"}},
}

expectedBulk = {
    "OK": True,
    "Value": {
        "BulkJobObject": "",
        1: {"a1": "aa1", "TransformationID": 1, "b1": "bb1", "Site": "MySite", "JobType": "User"},
        2: {"a2": "aa2", "TransformationID": 1, "b2": "bb2", "InputData": ["a1", "a2"], "JobType": "User"},
        3: {"TransformationID": 2, "a3": "aa3", "b3": "bb3", "JobType": "User"},
    },
}


@pytest.mark.parametrize(
    "taskDictionary, bulkSubmissionFlag, result, expectedRes",
    [
        (taskDict, False, True, expected),
        (taskDict, True, False, expectedBulk),
        (taskDictSimple, True, True, expectedBulk),
        (taskDictSimpleInputs, True, True, expectedBulkSimple),
        (taskDictNoInputs, True, False, expectedBulk),
        (taskDictNoInputsNoSite, True, True, expectedBulk),
    ],
)
def test_prepareTranformationTasks(taskDictionary, bulkSubmissionFlag, result, expectedRes):
    res = wfTasks.prepareTransformationTasks(
        "", taskDictionary, "test_user", "test_group", "test_DN", bulkSubmissionFlag=bulkSubmissionFlag
    )
    assert res["OK"] == result
    if res["OK"]:
        for key, value in res["Value"].items():
            if key != "BulkJobObject":
                assert key in expectedRes["Value"]
                for tKey, tValue in value.items():
                    assert tKey in expectedRes["Value"][key]
                    if tKey == "TaskObject" and tValue:
                        assert isinstance(tValue, Job)
                    else:
                        assert tValue == expectedRes["Value"][key][tKey]


def ourgetSitesForSE(ses):
    if ses == ["pippo"] or ses == "pippo":
        return {"OK": True, "Value": ["Site1"]}
    elif ses == ["pluto"] or ses == "pluto":
        return {"OK": True, "Value": ["Site2"]}
    elif ses == ["pippo", "pluto"] or ses == "pippo,pluto":
        return {"OK": True, "Value": ["Site1", "Site2"]}


@pytest.mark.parametrize(
    "paramsDict, expected",
    [
        ({"Site": "", "TargetSE": ""}, ["ANY"]),
        ({"Site": "ANY", "TargetSE": ""}, ["ANY"]),
        ({"TargetSE": "Unknown"}, ["ANY"]),
        ({"Site": "Site2", "TargetSE": ""}, ["Site2"]),
        ({"Site": "Site1;Site2", "TargetSE": "pippo"}, ["Site1"]),
        ({"Site": "Site1;Site2", "TargetSE": "pippo,pluto"}, ["Site1", "Site2"]),
        ({"Site": "Site1;Site2;Site3", "TargetSE": "pippo,pluto"}, ["Site1", "Site2"]),
        ({"Site": "Site2", "TargetSE": "pippo,pluto"}, ["Site2"]),
        ({"Site": "ANY", "TargetSE": "pippo,pluto"}, ["Site1", "Site2"]),
        ({"Site": "Site1", "TargetSE": "pluto"}, []),
    ],
)
def test__handleDestination(mocker, paramsDict, expected):
    mocker.patch("DIRAC.TransformationSystem.Client.TaskManagerPlugin.getSitesForSE", side_effect=ourgetSitesForSE)
    res = wfTasks._handleDestination(paramsDict)
    assert sorted(res) == sorted(expected)



"""
DummySyncPlugin
"""


class DummySyncPlugin(object):
    """Dummy Synchronization plugin that does nothing.
    It is used to document what are the requirements for a real Sync plugin.

    Such plugins are meant to validate user's information about to be added to the CS,
    or to complete it with various sources.
    They are called by the
    :py:meth:`DIRAC.ConfigurationSystem.Client.VOMS2CSSynchronizer.VOMS2CSSynchronizer.syncCSWithVOMS`
    """

    def __init__(self):
        """The constructor does not receive any argument.
        Note that the plugin is instanciated only once, so this is
        a good place to do global initialization.
        """
        pass

    def verifyAndUpdateUserInfo(self, username, userDict):
        """
          This method is expected to validate the user's data passed
          as parameter, but is also allowed to extend them.

          In case the validation was not to pass, this method must raise
          ``ValueError``. The user would then not be added to the CS.

        :param username: DIRAC name of the user to be added
        :param userDict: user information collected by the VOMS2CSAgent. Typical keys include
                         ``DN``,``CA``, ``Email``, ``Groups``
        :returns: None
        :raise ValueError: in case user information do not pass validation

        """
        raise NotImplementedError("Must be implemented by the plugin")



#!/usr/bin/env python
"""
Monitor the jobs present in the repository
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("RepoDir:  Location of Job Repository")
    _, args = Script.parseCommandLine(ignoreErrors=False)

    repoLocation = args[0]
    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac(withRepo=True, repoLocation=repoLocation)

    exitCode = 0
    result = dirac.monitorRepository(printOutput=True)
    if not result["OK"]:
        print("ERROR: ", result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



"""GSIFTP module based on the GFAL2_StorageBase class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# from DIRAC
from DIRAC.Resources.Storage.GFAL2_StorageBase import GFAL2_StorageBase
from DIRAC import gLogger
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse


class GFAL2_GSIFTPStorage(GFAL2_StorageBase):

    """.. class:: GFAL2_GSIFTPStorage

    GSIFTP interface to StorageElement using gfal2
    """

    _INPUT_PROTOCOLS = ["file", "gsiftp"]
    _OUTPUT_PROTOCOLS = ["gsiftp"]

    def __init__(self, storageName, parameters):
        """c'tor"""
        # # init base class
        super(GFAL2_GSIFTPStorage, self).__init__(storageName, parameters)
        self.srmSpecificParse = False

        self.log = gLogger.getSubLogger("GFAL2_GSIFTPStorage")

        self.pluginName = "GFAL2_GSIFTP"

        # We don't need extended attributes for metadata
        self._defaultExtendedAttributes = None

    def __addDoubleSlash(self, res):
        """Utilities to add the double slash between the host(:port) and the path

        :param res: DIRAC return structure which contains an URL if S_OK
        :return: DIRAC structure with corrected URL
        """
        if not res["OK"]:
            return res
        url = res["Value"]
        res = pfnparse(url, srmSpecific=self.srmSpecificParse)
        if not res["OK"]:
            return res
        urlDict = res["Value"]
        urlDict["Path"] = "/" + urlDict["Path"]
        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def getURLBase(self, withWSUrl=False):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_GSIFTPStorage, self).getURLBase(withWSUrl=withWSUrl))

    def constructURLFromLFN(self, lfn, withWSUrl=False):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_GSIFTPStorage, self).constructURLFromLFN(lfn=lfn, withWSUrl=withWSUrl))

    def getCurrentURL(self, fileName):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_GSIFTPStorage, self).getCurrentURL(fileName))



"""
Utilities to create replication transformations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC.TransformationSystem.Client.Transformation import Transformation
from DIRAC import gLogger, S_OK, S_ERROR


def createDataTransformation(
    flavour,
    targetSE,
    sourceSE,
    metaKey,
    metaValue,
    extraData=None,
    extraname="",
    groupSize=1,
    plugin="Broadcast",
    tGroup=None,
    tBody=None,
    enable=False,
):
    """Creates the replication transformation based on the given parameters.

    :param str flavour: Flavour of replication to create: Replication or Moving
    :param targetSE: Destination for files
    :type targetSE: python:list or str
    :param sourceSE: Origin of files
    :type sourceSE: python:list or str
    :param int metaKey: Meta key to identify input files
    :param int metaValue: Meta value to identify input files
    :param dict metaData: Additional meta data to use to identify input files
    :param str extraname: addition to the transformation name, only needed if the same transformation was already created
    :param int groupSize: number of files per transformation taks
    :param str plugin: plugin to use
    :param str tGroup: transformation group to set
    :param tBody: transformation body to set
    :param bool enable: if true submit the transformation, otherwise dry run
    :returns: S_OK (with the transformation object, if successfully added), S_ERROR
    """
    metadata = {metaKey: metaValue}
    if isinstance(extraData, dict):
        metadata.update(extraData)

    gLogger.debug("Using %r for metadata search" % metadata)

    if isinstance(targetSE, six.string_types):
        targetSE = [targetSE]

    if isinstance(sourceSE, (list, tuple)):
        sourceSE = "%s" % (",".join(sourceSE))

    gLogger.debug("Using plugin: %r" % plugin)

    if flavour not in ("Replication", "Moving"):
        return S_ERROR("Unsupported flavour %s" % flavour)

    transVerb = {"Replication": "Replicate", "Moving": "Move"}[flavour]
    transGroup = {"Replication": "Replication", "Moving": "Moving"}[flavour] if not tGroup else tGroup

    trans = Transformation()
    transName = "%s_%s_%s" % (transVerb, str(metaValue), ",".join(targetSE))
    if extraname:
        transName += "_%s" % extraname

    trans.setTransformationName(transName)
    description = "%s files for %s %s to %s" % (transVerb, metaKey, str(metaValue), ",".join(targetSE))
    trans.setDescription(description[:255])
    trans.setLongDescription(description)
    trans.setType("Replication")
    trans.setTransformationGroup(transGroup)
    trans.setGroupSize(groupSize)
    trans.setPlugin(plugin)

    transBody = (
        {
            "Moving": [
                ("ReplicateAndRegister", {"SourceSE": sourceSE, "TargetSE": targetSE}),
                ("RemoveReplica", {"TargetSE": sourceSE}),
            ],
            "Replication": "",  # empty body
        }[flavour]
        if tBody is None
        else tBody
    )

    trans.setBody(transBody)
    trans.setInputMetaQuery(metadata)

    if sourceSE:
        res = trans.setSourceSE(sourceSE)
        if not res["OK"]:
            return S_ERROR("SourceSE not valid: %s" % res["Message"])
    res = trans.setTargetSE(targetSE)
    if not res["OK"]:
        return S_ERROR("TargetSE not valid: %s" % res["Message"])

    if not enable:
        gLogger.always("Dry run, not creating transformation")
        return S_OK()

    res = trans.addTransformation()
    if not res["OK"]:
        return res
    gLogger.verbose(res)
    trans.setStatus("Active")
    trans.setAgentType("Automatic")

    gLogger.always("Successfully created replication transformation")
    return S_OK(trans)



"""
:mod: FileCatalogProxyHandler

.. module: FileCatalogProxyHandler
  :synopsis: This is a service which represents a DISET proxy to the File
  Catalog

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
import os
import six

# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Resources.Catalog.FileCatalogFactory import FileCatalogFactory
from DIRAC.Core.Utilities.Subprocess import pythonCall
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager

__RCSID__ = "$Id$"


def initializeFileCatalogProxyHandler(serviceInfo):
    """service initalisation"""
    return S_OK()


class FileCatalogProxyHandler(RequestHandler):
    """
    .. class:: FileCatalogProxyHandler
    """

    types_callProxyMethod = [list(six.string_types), list(six.string_types), [tuple, list], dict]

    def export_callProxyMethod(self, fcName, methodName, args, kargs):
        """A generic method to call methods of the Storage Element."""
        res = pythonCall(120, self.__proxyWrapper, fcName, methodName, args, kargs)
        if res["OK"]:
            return res["Value"]
        else:
            return res

    def __proxyWrapper(self, fcName, methodName, args, kwargs):
        """The wrapper will obtain the client proxy and set it up in the environment.
            The required functionality is then executed and returned to the client.

        :param self: self reference
        :param str name: fcn name
        :param tuple args: fcn args
        :param dict kwargs: fcn keyword args
        """
        result = self.__prepareSecurityDetails()
        if not result["OK"]:
            return result
        proxyLocation = result["Value"]
        try:
            result = FileCatalogFactory().createCatalog(fcName)
            if result["OK"]:
                fileCatalog = result["Value"]
                method = getattr(fileCatalog, methodName)
            else:
                return result
        except AttributeError as error:
            errStr = "%s proxy: no method named %s" % (fcName, methodName)
            gLogger.exception(errStr, methodName, error)
            return S_ERROR(errStr)
        try:
            result = method(*args, **kwargs)
            if os.path.exists(proxyLocation):
                os.remove(proxyLocation)
            return result
        except Exception as error:
            if os.path.exists(proxyLocation):
                os.remove(proxyLocation)
            errStr = "%s proxy: Exception while performing %s" % (fcName, methodName)
            gLogger.exception(errStr, error)
            return S_ERROR(errStr)

    def __prepareSecurityDetails(self, vomsFlag=True):
        """Obtains the connection details for the client"""
        try:
            credDict = self.getRemoteCredentials()
            clientDN = credDict["DN"]
            clientUsername = credDict["username"]
            clientGroup = credDict["group"]
            gLogger.debug("Getting proxy for %s@%s (%s)" % (clientUsername, clientGroup, clientDN))
            if vomsFlag:
                result = gProxyManager.downloadVOMSProxyToFile(clientDN, clientGroup)
            else:
                result = gProxyManager.downloadProxyToFile(clientDN, clientGroup)
            if not result["OK"]:
                return result
            gLogger.debug("Updating environment.")
            os.environ["X509_USER_PROXY"] = result["Value"]
            return result
        except Exception as error:
            exStr = "__getConnectionDetails: Failed to get client connection details."
            gLogger.exception(exStr, "", error)
            return S_ERROR(exStr)



"""Base class to give baseline for all the FCCondition plugins.
     It is important to note that in the FCConditionParser, the plugin is called for each
     and every lfn. This greatly simplifies the development of the plugin.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


class FCConditionBasePlugin(object):
    """Base class to give baseline for all the FCCondition plugins.
    It is important to note that in the FCConditionParser, the plugin is called for each
    and every lfn. This greatly simplifies the development of the plugin.
    """

    def __init__(self, conditions):
        """Gives the parameter of the evaluation to be done.
        The expression is defined as

          <PluginName> = <Plugin params>

        conditions is the string <Plugin params>
        They have to be interpreted
        """

        self.conditions = conditions

    def eval(self, **kwargs):
        """eval evaluates param of the plugins against all the info given
        and returns True or False.
        """
        raise NotImplementedError("to be implemented in the derived class")



""" Basic functions for interacting with CS objects
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import os

import DIRAC
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.Refresher import gRefresher

__RCSID__ = "$Id$"


class ConfigurationClient(object):
    def __init__(self, fileToLoadList=None):
        """C'or

        :param list fileToLoadList: files to load
        """
        self.diracConfigFilePath = os.path.join(DIRAC.rootPath, "etc", "dirac.cfg")
        if fileToLoadList and isinstance(fileToLoadList, list):
            for fileName in fileToLoadList:
                gConfigurationData.loadFile(fileName)

    def loadFile(self, fileName):
        """Load file

        :param str fileName: file name

        :return: S_OK()/S_ERROR()
        """
        return gConfigurationData.loadFile(fileName)

    def loadCFG(self, cfg):
        """Load CFG

        :param CFG() cfg: CFG object

        :return: S_OK()/S_ERROR()
        """
        return gConfigurationData.mergeWithLocal(cfg)

    def forceRefresh(self, fromMaster=False):
        """Force refresh

        :param bool fromMaster: refresh from master

        :return: S_OK()/S_ERROR()
        """
        return gRefresher.forceRefresh(fromMaster=fromMaster)

    def dumpLocalCFGToFile(self, fileName):
        """Dump local configuration to file

        :param str fileName: file name

        :return: S_OK()/S_ERROR()
        """
        return gConfigurationData.dumpLocalCFGToFile(fileName)

    def dumpRemoteCFGToFile(self, fileName):
        """Dump remote configuration to file

        :param str fileName: file name

        :return: S_OK()/S_ERROR()
        """
        return gConfigurationData.dumpRemoteCFGToFile(fileName)

    def addListenerToNewVersionEvent(self, functor):
        """Add listener to new version event

        :param str functor: functor
        """
        gRefresher.addListenerToNewVersionEvent(functor)

    def dumpCFGAsLocalCache(self, fileName=None, raw=False):
        """Dump local CFG cache to file

        :param str fileName: file name
        :param bool raw: raw

        :return: S_OK(str)/S_ERROR()
        """
        cfg = gConfigurationData.mergedCFG.clone()
        try:
            if not raw and cfg.isSection("DIRAC"):
                diracSec = cfg["DIRAC"]
                if diracSec.isSection("Configuration"):  # pylint: disable=no-member
                    confSec = diracSec["Configuration"]  # pylint: disable=unsubscriptable-object
                    for opt in ("Servers", "MasterServer"):
                        if confSec.isOption(opt):
                            confSec.deleteKey(opt)
            strData = str(cfg)
            if fileName:
                with open(fileName, "w") as fd:
                    fd.write(strData)
        except Exception as e:
            return S_ERROR("Can't write to file %s: %s" % (fileName, str(e)))
        return S_OK(strData)

    def getServersList(self):
        """Get list of servers

        :return: list
        """
        return gConfigurationData.getServers()

    def useServerCertificate(self):
        """Get using server certificate status

        :return: bool
        """
        return gConfigurationData.useServerCertificate()

    def getValue(self, optionPath, defaultValue=None):
        """Get configuration value

        :param str optionPath: option path
        :param defaultValue: default value

        :return: type(defaultValue)
        """
        retVal = self.getOption(optionPath, defaultValue)
        return retVal["Value"] if retVal["OK"] else defaultValue

    def getOption(self, optionPath, typeValue=None):
        """Get configuration option

        :param str optionPath: option path
        :param typeValue: type of value

        :return: S_OK()/S_ERROR()
        """
        gRefresher.refreshConfigurationIfNeeded()
        optionValue = gConfigurationData.extractOptionFromCFG(optionPath)

        if optionValue is None:
            return S_ERROR(
                "Path %s does not exist or it's not an option" % optionPath,
                callStack=["ConfigurationClient.getOption"],
            )

        # Value has been returned from the configuration
        if typeValue is None:
            return S_OK(optionValue)

        # Casting to typeValue's type
        if not isinstance(typeValue, type):
            # typeValue is not a type but a default object
            requestedType = type(typeValue)
        else:
            requestedType = typeValue

        if requestedType in (list, tuple, set):
            try:
                return S_OK(requestedType(List.fromChar(optionValue, ",")))
            except Exception as e:
                return S_ERROR("Can't convert value (%s) to comma separated list \n%s" % (str(optionValue), repr(e)))
        elif requestedType == bool:
            try:
                return S_OK(optionValue.lower() in ("y", "yes", "true", "1"))
            except Exception as e:
                return S_ERROR("Can't convert value (%s) to Boolean \n%s" % (str(optionValue), repr(e)))
        elif requestedType == dict:
            try:
                splitOption = List.fromChar(optionValue, ",")
                value = {}
                for opt in splitOption:
                    keyVal = [x.strip() for x in opt.split(":")]
                    if len(keyVal) == 1:
                        keyVal.append(True)
                    value[keyVal[0]] = keyVal[1]
                return S_OK(value)
            except Exception as e:
                return S_ERROR("Can't convert value (%s) to Dict \n%s" % (str(optionValue), repr(e)))
        else:
            try:
                return S_OK(requestedType(optionValue))
            except Exception as e:
                return S_ERROR(
                    "Type mismatch between default (%s) and configured value (%s) \n%s"
                    % (str(typeValue), optionValue, repr(e))
                )

    def getSections(self, sectionPath, listOrdered=True):
        """Get configuration sections

        :param str sectionPath: section path
        :param bool listOrdered: ordered

        :return: S_OK(list)/S_ERROR()
        """
        gRefresher.refreshConfigurationIfNeeded()
        sectionList = gConfigurationData.getSectionsFromCFG(sectionPath, ordered=listOrdered)
        if isinstance(sectionList, list):
            return S_OK(sectionList)
        else:
            return S_ERROR("Path %s does not exist or it's not a section" % sectionPath)

    def getOptions(self, sectionPath, listOrdered=True):
        """Get configuration options

        :param str sectionPath: section path
        :param bool listOrdered: ordered

        :return: S_OK(list)/S_ERROR()
        """
        gRefresher.refreshConfigurationIfNeeded()
        optionList = gConfigurationData.getOptionsFromCFG(sectionPath, ordered=listOrdered)
        if isinstance(optionList, list):
            return S_OK(optionList)
        else:
            return S_ERROR("Path %s does not exist or it's not a section" % sectionPath)

    def getOptionsDict(self, sectionPath):
        """Get configuration options in dictionary

        :param str sectionPath: section path

        :return: S_OK(dict)/S_ERROR()
        """
        gRefresher.refreshConfigurationIfNeeded()
        optionsDict = {}
        optionList = gConfigurationData.getOptionsFromCFG(sectionPath)
        if isinstance(optionList, list):
            for option in optionList:
                optionsDict[option] = gConfigurationData.extractOptionFromCFG("%s/%s" % (sectionPath, option))
            return S_OK(optionsDict)
        else:
            return S_ERROR("Path %s does not exist or it's not a section" % sectionPath)

    def getOptionsDictRecursively(self, sectionPath):
        """Get configuration options in dictionary recursively

        :param str sectionPath: section path

        :return: S_OK(dict)/S_ERROR()
        """
        if not gConfigurationData.mergedCFG.isSection(sectionPath):
            return S_ERROR("Path %s does not exist or it's not a section" % sectionPath)
        return S_OK(gConfigurationData.mergedCFG.getAsDict(sectionPath))

    def getConfigurationTree(self, root="", *filters):
        """Create a dictionary with all sections, subsections and options
        starting from given root. Result can be filtered.

        :param str root: Starting point in the configuration tree.
        :param filters: Select results that contain given substrings (check full path, i.e. with option name)
        :type filters: str or python:list[str]

        :return: S_OK(dict)/S_ERROR() -- dictionary where keys are paths taken from
                 the configuration (e.g. /Systems/Configuration/...).
                 Value is "None" when path points to a section
                 or not "None" if path points to an option.
        """

        # check if root is an option (special case)
        option = self.getOption(root)
        if option["OK"]:
            result = {root: option["Value"]}

        else:
            result = {root: None}
            for substr in filters:
                if substr not in root:
                    result = {}
                    break

            # remove slashes at the end
            root = root.rstrip("/")

            # get options of current root
            options = self.getOptionsDict(root)
            if not options["OK"]:
                return S_ERROR("getOptionsDict() failed with message: %s" % options["Message"])

            for key, value in options["Value"].items():
                path = cfgPath(root, key)
                addOption = True
                for substr in filters:
                    if substr not in path:
                        addOption = False
                        break

                if addOption:
                    result[path] = value

            # get subsections of the root
            sections = self.getSections(root)
            if not sections["OK"]:
                return S_ERROR("getSections() failed with message: %s" % sections["Message"])

            # recursively go through subsections and get their subsections
            for section in sections["Value"]:
                subtree = self.getConfigurationTree("%s/%s" % (root, section), *filters)
                if not subtree["OK"]:
                    return S_ERROR("getConfigurationTree() failed with message: %s" % sections["Message"])
                result.update(subtree["Value"])

        return S_OK(result)

    def setOptionValue(self, optionPath, value):
        """Set a value in the local configuration

        :param str optionPath: option path
        :param str value: value
        """
        gConfigurationData.setOptionInCFG(optionPath, value)



"""
Test Config for External Libraries:

- External libs may use the standard logging library with a different configuration
- We want to make sure this does not affect the DIRAC logging system based on it

"""

__RCSID__ = "$Id$"

import logging
from six import StringIO
import pytest

from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import gLogger, cleaningLog, gLoggerReset


@pytest.mark.parametrize(
    "isEnabled, loggingLevel, numberHandlers, handlerType",
    [
        (True, logging.DEBUG, 1, logging.StreamHandler),
        (False, logging.DEBUG, 1, logging.NullHandler),
    ],
)
def test_logsFromExtLibsHandlers(isEnabled, loggingLevel, numberHandlers, handlerType):
    """
    Check that handlers attached are different according to the value of enableLogsFromExternalLibs()
    """
    gLoggerReset()
    # when enable, Logging should also report logs from external library
    # a StreamHandler should be attached to the root logger by default
    if isEnabled:
        gLogger.enableLogsFromExternalLibs()
    # in the other case, Logging should not report logs from external library
    # a NullHandler disables the emission of the log records going to the root logger
    else:
        gLogger.disableLogsFromExternalLibs()

    handlers = logging.getLogger().handlers

    assert logging.getLogger().getEffectiveLevel() == loggingLevel

    assert numberHandlers == len(handlers)
    assert isinstance(handlers[0], handlerType)


@pytest.mark.parametrize(
    "isEnabled, loggerName, message, expected",
    [
        (True, "", "message", "ExternalLibrary/root INFO: message\n"),
        (True, "sublog", "message", "ExternalLibrary/sublog INFO: message\n"),
        (False, "", "message", ""),
        (False, "sublog", "message", ""),
    ],
)
def test_logsFromExtLibsLogs(isEnabled, loggerName, message, expected):
    """
    Check whether logs are displayed according to the value of enableLogsFromExternalLibs()
    """
    gLoggerReset()
    # when enable, Logging should also report logs from external library
    # logs from external libs should appear
    if isEnabled:
        gLogger.enableLogsFromExternalLibs()
    # in the other case, Logging should not report logs from external library
    # logs from external libs shouldn't appear
    else:
        gLogger.disableLogsFromExternalLibs()

    # modify the output to capture logs of the root logger
    bufferRoot = StringIO()
    logging.getLogger().handlers[0].stream = bufferRoot

    logging.getLogger(loggerName).info(message)
    logstring = cleaningLog(bufferRoot.getvalue())

    assert expected == logstring


def test_logsFromExtLibsPropag():
    """
    Test the no propagation of the logs from the Logging objects to the root logger of 'logging'
    """
    capturedBackend, _, _ = gLoggerReset()
    gLogger.enableLogsFromExternalLibs()

    # modify the output to capture logs of the root logger
    bufferRoot = StringIO()
    logging.getLogger().handlers[0].stream = bufferRoot

    gLogger.error("message")

    assert capturedBackend.getvalue() != ""
    assert bufferRoot.getvalue() == ""


def test_logsFromExtLibsMultCalls():
    """
    Calls the method several times to see whether we have duplication of the logs
    """
    gLoggerReset()
    for i in range(5):
        gLogger.enableLogsFromExternalLibs()
    handlers = logging.getLogger().handlers

    assert len(handlers) == 1
    assert isinstance(handlers[0], logging.StreamHandler)

    for i in range(5):
        gLogger.disableLogsFromExternalLibs()
    handlers = logging.getLogger().handlers

    assert len(handlers) == 1
    assert isinstance(handlers[0], logging.NullHandler)

    gLogger.enableLogsFromExternalLibs()
    gLogger.disableLogsFromExternalLibs()
    gLogger.enableLogsFromExternalLibs()
    gLogger.disableLogsFromExternalLibs()
    handlers = logging.getLogger().handlers

    assert len(handlers) == 1
    assert isinstance(handlers[0], logging.NullHandler)

    gLogger.disableLogsFromExternalLibs()
    gLogger.enableLogsFromExternalLibs()
    gLogger.disableLogsFromExternalLibs()
    gLogger.enableLogsFromExternalLibs()
    handlers = logging.getLogger().handlers

    assert len(handlers) == 1
    assert isinstance(handlers[0], logging.StreamHandler)



""" FileReport module defines the FileReport class, to report file status to the transformation DB
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

from DIRAC import S_OK
from DIRAC.Core.Utilities import DEncode
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.RequestManagementSystem.Client.Operation import Operation

__RCSID__ = "$Id$"


class FileReport(object):
    """A stateful object for reporting to TransformationDB"""

    def __init__(self, server="Transformation/TransformationManager"):
        """c'tor

        self.transClient is a TransformationClient object
        """
        self.transClient = TransformationClient()
        self.transClient.setServer(server)
        self.statusDict = {}
        self.transformation = None
        self.force = False

    def setFileStatus(self, transformation, lfn, status, sendFlag=False):
        """Set file status in the context of the given transformation"""
        if not self.transformation:
            self.transformation = transformation
        if isinstance(lfn, (list, dict, tuple)):
            self.statusDict.update(dict.fromkeys(lfn, status))
        else:
            self.statusDict[lfn] = status
        if sendFlag:
            return self.commit()
        return S_OK()

    def setCommonStatus(self, status):
        """Set common status for all files in the internal cache"""
        for lfn in self.statusDict:
            self.statusDict[lfn] = status
        return S_OK()

    def getFiles(self):
        """Get the statuses of the files already accumulated in the FileReport object"""
        return copy.deepcopy(self.statusDict)

    def commit(self):
        """Commit pending file status update records"""
        if not self.statusDict:
            return S_OK({})

        result = self.transClient.setFileStatusForTransformation(self.transformation, self.statusDict, force=self.force)
        if result["OK"]:
            self.statusDict = {}
        return result

    def generateForwardDISET(self):
        """Commit the accumulated records and generate request eventually"""
        result = self.commit()
        commitOp = None
        if not result["OK"]:
            # Generate Request
            commitOp = Operation()
            commitOp.Type = "SetFileStatus"
            commitOp.Arguments = DEncode.encode(
                {"transformation": self.transformation, "statusDict": self.statusDict, "force": self.force}
            )

        return S_OK(commitOp)



""" ProxyManager is the implementation of the ProxyManagement service in the DISET framework

    .. literalinclude:: ../ConfigTemplate.cfg
      :start-after: ##BEGIN ProxyManager:
      :end-before: ##END
      :dedent: 2
      :caption: ProxyManager options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Security import Properties
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers import Registry


DEFAULT_MAIL_FROM = "proxymanager@diracgrid.org"


class ProxyManagerHandler(RequestHandler):

    __maxExtraLifeFactor = 1.5
    __proxyDB = None

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        useMyProxy = cls.srv_getCSOption("UseMyProxy", False)
        mailFrom = getServiceOption(serviceInfoDict, "MailFrom", DEFAULT_MAIL_FROM)

        try:
            result = ObjectLoader().loadObject("FrameworkSystem.DB.ProxyDB")
            if not result["OK"]:
                gLogger.error("Failed to load ProxyDB class: %s" % result["Message"])
                return result
            dbClass = result["Value"]

            cls.__proxyDB = dbClass(useMyProxy=useMyProxy, mailFrom=mailFrom)

        except RuntimeError as excp:
            return S_ERROR("Can't connect to ProxyDB: %s" % excp)
        gThreadScheduler.addPeriodicTask(900, cls.__proxyDB.purgeExpiredTokens, elapsedTime=900)
        gThreadScheduler.addPeriodicTask(900, cls.__proxyDB.purgeExpiredRequests, elapsedTime=900)
        gThreadScheduler.addPeriodicTask(21600, cls.__proxyDB.purgeLogs)
        gThreadScheduler.addPeriodicTask(3600, cls.__proxyDB.purgeExpiredProxies)
        if useMyProxy:
            gLogger.info("MyProxy: %s\n MyProxy Server: %s" % (useMyProxy, cls.__proxyDB.getMyProxyServer()))
        return S_OK()

    def __generateUserProxiesInfo(self):
        """Generate information dict about user proxies

        :return: dict
        """
        proxiesInfo = {}
        credDict = self.getRemoteCredentials()
        result = Registry.getDNForUsername(credDict["username"])
        if not result["OK"]:
            return result
        selDict = {"UserDN": result["Value"]}
        result = self.__proxyDB.getProxiesContent(selDict, {})
        if not result["OK"]:
            return result
        contents = result["Value"]
        userDNIndex = contents["ParameterNames"].index("UserDN")
        userGroupIndex = contents["ParameterNames"].index("UserGroup")
        expirationIndex = contents["ParameterNames"].index("ExpirationTime")
        for record in contents["Records"]:
            userDN = record[userDNIndex]
            if userDN not in proxiesInfo:
                proxiesInfo[userDN] = {}
            userGroup = record[userGroupIndex]
            proxiesInfo[userDN][userGroup] = record[expirationIndex]
        return proxiesInfo

    def __addKnownUserProxiesInfo(self, retDict):
        """Given a S_OK/S_ERR add a proxies entry with info of all the proxies a user has uploaded

        :return: S_OK(dict)/S_ERROR()
        """
        retDict["proxies"] = self.__generateUserProxiesInfo()
        return retDict

    auth_getUserProxiesInfo = ["authenticated"]
    types_getUserProxiesInfo = []

    def export_getUserProxiesInfo(self):
        """Get the info about the user proxies in the system

        :return: S_OK(dict)
        """
        return S_OK(self.__generateUserProxiesInfo())

    # WARN: Since v7r1 requestDelegationUpload method use only first argument!
    # WARN:   Second argument for compatibility with older versions
    types_requestDelegationUpload = [six.integer_types]

    def export_requestDelegationUpload(self, requestedUploadTime, diracGroup=None):
        """Request a delegation. Send a delegation request to client

        :param int requestedUploadTime: requested live time

        :return: S_OK(dict)/S_ERROR() -- dict contain id and proxy as string of the request
        """
        if diracGroup:
            self.log.warn("Since v7r1 requestDelegationUpload method use only first argument!")
        credDict = self.getRemoteCredentials()
        user = "%s:%s" % (credDict["username"], credDict["group"])
        result = self.__proxyDB.generateDelegationRequest(credDict["x509Chain"], credDict["DN"])
        if result["OK"]:
            gLogger.info("Upload request by %s given id %s" % (user, result["Value"]["id"]))
        else:
            gLogger.error("Upload request failed", "by %s : %s" % (user, result["Message"]))
        return result

    types_completeDelegationUpload = [six.integer_types, six.string_types]

    def export_completeDelegationUpload(self, requestId, pemChain):
        """Upload result of delegation

        :param int requestId: identity number
        :param str pemChain: certificate as string

        :return: S_OK(dict)/S_ERROR() -- dict contain proxies
        """
        credDict = self.getRemoteCredentials()
        userId = "%s:%s" % (credDict["username"], credDict["group"])
        retVal = self.__proxyDB.completeDelegation(requestId, credDict["DN"], pemChain)
        if not retVal["OK"]:
            gLogger.error("Upload proxy failed", "id: %s user: %s message: %s" % (requestId, userId, retVal["Message"]))
            return self.__addKnownUserProxiesInfo(retVal)
        gLogger.info("Upload %s by %s completed" % (requestId, userId))
        return self.__addKnownUserProxiesInfo(S_OK())

    types_getRegisteredUsers = []

    def export_getRegisteredUsers(self, validSecondsRequired=0):
        """Get the list of users who have a valid proxy in the system

        :param int validSecondsRequired: required seconds the proxy is valid for

        :return: S_OK(list)/S_ERROR() -- list contain dicts with user name, DN, group
                                         expiration time, persistent flag
        """
        credDict = self.getRemoteCredentials()
        if Properties.PROXY_MANAGEMENT not in credDict["properties"]:
            return self.__proxyDB.getUsers(validSecondsRequired, userMask=credDict["username"])
        return self.__proxyDB.getUsers(validSecondsRequired)

    def __checkProperties(self, requestedUserDN, requestedUserGroup):
        """Check the properties and return if they can only download limited proxies if authorized

        :param str requestedUserDN: user DN
        :param str requestedUserGroup: DIRAC group

        :return: S_OK(boolean)/S_ERROR()
        """
        credDict = self.getRemoteCredentials()
        if Properties.FULL_DELEGATION in credDict["properties"]:
            return S_OK(False)
        if Properties.LIMITED_DELEGATION in credDict["properties"]:
            return S_OK(True)
        if Properties.PRIVATE_LIMITED_DELEGATION in credDict["properties"]:
            if credDict["DN"] != requestedUserDN:
                return S_ERROR("You are not allowed to download any proxy")
            if Properties.PRIVATE_LIMITED_DELEGATION not in Registry.getPropertiesForGroup(requestedUserGroup):
                return S_ERROR("You can't download proxies for that group")
            return S_OK(True)
        # Not authorized!
        return S_ERROR("You can't get proxies!")

    types_getProxy = [six.string_types, six.string_types, six.string_types, six.integer_types]

    def export_getProxy(self, userDN, userGroup, requestPem, requiredLifetime):
        """Get a proxy for a userDN/userGroup

        :param requestPem: PEM encoded request object for delegation
        :param requiredLifetime: Argument for length of proxy

          * Properties:
              * FullDelegation <- permits full delegation of proxies
              * LimitedDelegation <- permits downloading only limited proxies
              * PrivateLimitedDelegation <- permits downloading only limited proxies for one self
        """
        credDict = self.getRemoteCredentials()

        result = self.__checkProperties(userDN, userGroup)
        if not result["OK"]:
            return result
        forceLimited = result["Value"]

        self.__proxyDB.logAction("download proxy", credDict["DN"], credDict["group"], userDN, userGroup)
        return self.__getProxy(userDN, userGroup, requestPem, requiredLifetime, forceLimited)

    def __getProxy(self, userDN, userGroup, requestPem, requiredLifetime, forceLimited):
        """Internal to get a proxy

        :param str userDN: user DN
        :param str userGroup: DIRAC group
        :param str requestPem: dump of request certificate
        :param int requiredLifetime: requested live time of proxy
        :param boolean forceLimited: limited proxy

        :return: S_OK(str)/S_ERROR()
        """
        retVal = self.__proxyDB.getProxy(userDN, userGroup, requiredLifeTime=requiredLifetime)
        if not retVal["OK"]:
            return retVal
        chain, secsLeft = retVal["Value"]
        # If possible we return a proxy 1.5 longer than requested
        requiredLifetime = int(min(secsLeft, requiredLifetime * self.__maxExtraLifeFactor))
        retVal = chain.generateChainFromRequestString(
            requestPem, lifetime=requiredLifetime, requireLimited=forceLimited
        )
        if not retVal["OK"]:
            return retVal
        return S_OK(retVal["Value"])

    types_getVOMSProxy = [
        six.string_types,
        six.string_types,
        six.string_types,
        six.integer_types,
        [six.string_types, type(None), bool],
    ]

    def export_getVOMSProxy(self, userDN, userGroup, requestPem, requiredLifetime, vomsAttribute=None):
        """Get a proxy for a userDN/userGroup

        :param requestPem: PEM encoded request object for delegation
        :param requiredLifetime: Argument for length of proxy
        :param vomsAttribute: VOMS attr to add to the proxy

          * Properties :
              * FullDelegation <- permits full delegation of proxies
              * LimitedDelegation <- permits downloading only limited proxies
              * PrivateLimitedDelegation <- permits downloading only limited proxies for one self
        """
        credDict = self.getRemoteCredentials()

        result = self.__checkProperties(userDN, userGroup)
        if not result["OK"]:
            return result
        forceLimited = result["Value"]

        self.__proxyDB.logAction("download voms proxy", credDict["DN"], credDict["group"], userDN, userGroup)
        return self.__getVOMSProxy(userDN, userGroup, requestPem, requiredLifetime, vomsAttribute, forceLimited)

    def __getVOMSProxy(self, userDN, userGroup, requestPem, requiredLifetime, vomsAttribute, forceLimited):
        retVal = self.__proxyDB.getVOMSProxy(
            userDN, userGroup, requiredLifeTime=requiredLifetime, requestedVOMSAttr=vomsAttribute
        )
        if not retVal["OK"]:
            return retVal
        chain, secsLeft = retVal["Value"]
        # If possible we return a proxy 1.5 longer than requested
        requiredLifetime = int(min(secsLeft, requiredLifetime * self.__maxExtraLifeFactor))
        return chain.generateChainFromRequestString(requestPem, lifetime=requiredLifetime, requireLimited=forceLimited)

    types_setPersistency = [six.string_types, six.string_types, bool]

    def export_setPersistency(self, userDN, userGroup, persistentFlag):
        """Set the persistency for a given dn/group

        :param str userDN: user DN
        :param str userGroup: DIRAC group
        :param boolean persistentFlag: if proxy persistent

        :return: S_OK()/S_ERROR()
        """
        retVal = self.__proxyDB.setPersistencyFlag(userDN, userGroup, persistentFlag)
        if not retVal["OK"]:
            return retVal
        credDict = self.getRemoteCredentials()
        self.__proxyDB.logAction(
            "set persistency to %s" % bool(persistentFlag), credDict["DN"], credDict["group"], userDN, userGroup
        )
        return S_OK()

    types_deleteProxyBundle = [(list, tuple)]

    def export_deleteProxyBundle(self, idList):
        """delete a list of id's

        :param list,tuple idList: list of identity numbers

        :return: S_OK(int)/S_ERROR()
        """
        errorInDelete = []
        deleted = 0
        for _id in idList:
            if len(_id) != 2:
                errorInDelete.append("%s doesn't have two fields" % str(_id))
            retVal = self.export_deleteProxy(_id[0], _id[1])
            if not retVal["OK"]:
                errorInDelete.append("%s : %s" % (str(_id), retVal["Message"]))
            else:
                deleted += 1
        if errorInDelete:
            return S_ERROR("Could not delete some proxies: %s" % ",".join(errorInDelete))
        return S_OK(deleted)

    types_deleteProxy = [(list, tuple)]

    def export_deleteProxy(self, userDN, userGroup):
        """Delete a proxy from the DB

        :param str userDN: user DN
        :param str userGroup: DIRAC group

        :return: S_OK()/S_ERROR()
        """
        credDict = self.getRemoteCredentials()
        if Properties.PROXY_MANAGEMENT not in credDict["properties"]:
            if userDN != credDict["DN"]:
                return S_ERROR("You aren't allowed!")
        retVal = self.__proxyDB.deleteProxy(userDN, userGroup)
        if not retVal["OK"]:
            return retVal
        self.__proxyDB.logAction("delete proxy", credDict["DN"], credDict["group"], userDN, userGroup)
        return S_OK()

    types_getContents = [dict, (list, tuple), six.integer_types, six.integer_types]

    def export_getContents(self, selDict, sortDict, start, limit):
        """Retrieve the contents of the DB

        :param dict selDict: selection fields
        :param list,tuple sortDict: sorting fields
        :param int start: search limit start
        :param int start: search limit amount

        :return: S_OK(dict)/S_ERROR() -- dict contain fields, record list, total records
        """
        credDict = self.getRemoteCredentials()
        if Properties.PROXY_MANAGEMENT not in credDict["properties"]:
            selDict["UserName"] = credDict["username"]
        return self.__proxyDB.getProxiesContent(selDict, sortDict, start, limit)

    types_getLogContents = [dict, (list, tuple), six.integer_types, six.integer_types]

    def export_getLogContents(self, selDict, sortDict, start, limit):
        """Retrieve the contents of the DB

        :param dict selDict: selection fields
        :param list,tuple sortDict: search filter
        :param int start: search limit start
        :param int start: search limit amount

        :return: S_OK(dict)/S_ERROR() -- dict contain fields, record list, total records
        """
        return self.__proxyDB.getLogsContent(selDict, sortDict, start, limit)

    types_generateToken = [six.string_types, six.string_types, six.integer_types]

    def export_generateToken(self, requesterDN, requesterGroup, tokenUses):
        """Generate tokens for proxy retrieval

        :param str requesterDN: user DN
        :param str requesterGroup: DIRAC group
        :param int tokenUses: number of uses

        :return: S_OK(tuple)/S_ERROR() -- tuple contain token, number uses
        """
        credDict = self.getRemoteCredentials()
        self.__proxyDB.logAction("generate tokens", credDict["DN"], credDict["group"], requesterDN, requesterGroup)
        return self.__proxyDB.generateToken(requesterDN, requesterGroup, numUses=tokenUses)

    types_getProxyWithToken = [
        six.string_types,
        six.string_types,
        six.string_types,
        six.integer_types,
        six.string_types,
    ]

    def export_getProxyWithToken(self, userDN, userGroup, requestPem, requiredLifetime, token):
        """Get a proxy for a userDN/userGroup

        :param requestPem: PEM encoded request object for delegation
        :param requiredLifetime: Argument for length of proxy
        :param token: Valid token to get a proxy

          * Properties:
              * FullDelegation <- permits full delegation of proxies
              * LimitedDelegation <- permits downloading only limited proxies
              * PrivateLimitedDelegation <- permits downloading only limited proxies for one self
        """
        credDict = self.getRemoteCredentials()
        result = self.__proxyDB.useToken(token, credDict["DN"], credDict["group"])
        gLogger.info("Trying to use token %s by %s:%s" % (token, credDict["DN"], credDict["group"]))
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Proxy token is invalid")
        self.__proxyDB.logAction("used token", credDict["DN"], credDict["group"], userDN, userGroup)

        result = self.__checkProperties(userDN, userGroup)
        if not result["OK"]:
            return result
        self.__proxyDB.logAction("download proxy with token", credDict["DN"], credDict["group"], userDN, userGroup)
        return self.__getProxy(userDN, userGroup, requestPem, requiredLifetime, True)

    types_getVOMSProxyWithToken = [
        six.string_types,
        six.string_types,
        six.string_types,
        six.integer_types,
        [six.string_types, type(None)],
    ]

    def export_getVOMSProxyWithToken(self, userDN, userGroup, requestPem, requiredLifetime, token, vomsAttribute=None):
        """Get a proxy for a userDN/userGroup

        :param requestPem: PEM encoded request object for delegation
        :param requiredLifetime: Argument for length of proxy
        :param vomsAttribute: VOMS attr to add to the proxy

          * Properties :
              * FullDelegation <- permits full delegation of proxies
              * LimitedDelegation <- permits downloading only limited proxies
              * PrivateLimitedDelegation <- permits downloading only limited proxies for one self
        """
        credDict = self.getRemoteCredentials()
        result = self.__proxyDB.useToken(token, credDict["DN"], credDict["group"])
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Proxy token is invalid")
        self.__proxyDB.logAction("used token", credDict["DN"], credDict["group"], userDN, userGroup)

        result = self.__checkProperties(userDN, userGroup)
        if not result["OK"]:
            return result
        self.__proxyDB.logAction("download voms proxy with token", credDict["DN"], credDict["group"], userDN, userGroup)
        return self.__getVOMSProxy(userDN, userGroup, requestPem, requiredLifetime, vomsAttribute, True)



#!/usr/bin/env python
########################################################################
# File :   dirac_admin_add_shifter
# Author : Federico Stagni
########################################################################
"""
Adds or modify a shifter, in the operations section of the CS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC import exit as DIRACExit, gLogger


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("ShifterRole:  Name of the shifter role, e.g. DataManager")
    Script.registerArgument("UserName:     A user name, as registered in Registry section")
    Script.registerArgument("DIRACGroup:   DIRAC Group, e.g. diracAdmin (the user has to have this role)")
    Script.parseCommandLine(ignoreErrors=True)

    csAPI = CSAPI()

    shifterRole, userName, diracGroup = Script.getPositionalArgs(group=True)
    res = csAPI.addShifter({shifterRole: {"User": userName, "Group": diracGroup}})
    if not res["OK"]:
        gLogger.error("Could not add shifter", ": " + res["Message"])
        DIRACExit(1)
    res = csAPI.commit()
    if not res["OK"]:
        gLogger.error("Could not add shifter", ": " + res["Message"])
        DIRACExit(1)
    gLogger.notice("Added shifter %s as user %s with group %s" % (shifterRole, userName, diracGroup))


if __name__ == "__main__":
    main()



"""
   DIRAC return dictionary

   Message values are converted to string

   keys are converted to string
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import sys
import traceback

import six

from DIRAC.Core.Utilities.DErrno import strerror


def S_ERROR(*args, **kwargs):
    """return value on error condition

    Arguments are either Errno and ErrorMessage or just ErrorMessage fro backward compatibility

    :param int errno: Error number
    :param string message: Error message
    :param list callStack: Manually override the CallStack attribute better performance
    """
    callStack = kwargs.pop("callStack", None)

    result = {"OK": False, "Errno": 0, "Message": ""}

    message = ""
    if args:
        if isinstance(args[0], six.integer_types):
            result["Errno"] = args[0]
            if len(args) > 1:
                message = args[1]
        else:
            message = args[0]

    if result["Errno"]:
        message = "%s ( %s : %s)" % (strerror(result["Errno"]), result["Errno"], message)
    result["Message"] = message

    if callStack is None:
        try:
            callStack = traceback.format_stack()
            callStack.pop()
        except Exception:
            callStack = []

    result["CallStack"] = callStack

    # print "AT >>> S_ERROR", result['OK'], result['Errno'], result['Message']
    # for item in result['CallStack']:
    #  print item

    return result


def S_OK(value=None):
    """return value on success

    :param value: value of the 'Value'
    :return: dictionary { 'OK' : True, 'Value' : value }
    """
    return {"OK": True, "Value": value}


def isReturnStructure(unk):
    """Check if value is an `S_OK`/`S_ERROR` object"""
    if not isinstance(unk, dict):
        return False
    if "OK" not in unk:
        return False
    if unk["OK"]:
        return "Value" in unk
    else:
        return "Message" in unk


def isSError(value):
    """Check if value is an `S_ERROR` object"""
    if not isinstance(value, dict):
        return False
    if "OK" not in value:
        return False
    return "Message" in value


def reprReturnErrorStructure(struct, full=False):
    errorNumber = struct.get("Errno", 0)
    message = struct.get("Message", "")
    if errorNumber:
        reprStr = "%s ( %s : %s)" % (strerror(errorNumber), errorNumber, message)
    else:
        reprStr = message

    if full:
        callStack = struct.get("CallStack")
        if callStack:
            reprStr += "\n" + "".join(callStack)

    return reprStr


def returnSingleResult(dictRes):
    """Transform the S_OK{Successful/Failed} dictionary convention into
    an S_OK/S_ERROR return. To be used when a single returned entity
    is expected from a generally bulk call.

    :param dictRes: S_ERROR or S_OK( "Failed" : {}, "Successful" : {})
    :returns: S_ERROR or S_OK(value)

    The following rules are applied:

    - if dictRes is an S_ERROR: returns it as is
    - we start by looking at the Failed directory
    - if there are several items in a dictionary, we return the first one
    - if both dictionaries are empty, we return S_ERROR
    - For an item in Failed, we return S_ERROR
    - Far an item in Successful we return S_OK

    Behavior examples (would be perfect unit test :-) )::

      {'Message': 'Kaput', 'OK': False} -> {'Message': 'Kaput', 'OK': False}
      {'OK': True, 'Value': {'Successful': {}, 'Failed': {'a': 1}}} -> {'Message': '1', 'OK': False}
      {'OK': True, 'Value': {'Successful': {'b': 2}, 'Failed': {}}} -> {'OK': True, 'Value': 2}
      {'OK': True, 'Value': {'Successful': {'b': 2}, 'Failed': {'a': 1}}} -> {'Message': '1', 'OK': False}
      {'OK': True, 'Value': {'Successful': {'b': 2}, 'Failed': {'a': 1, 'c': 3}}} -> {'Message': '1', 'OK': False}
      {'OK': True, 'Value': {'Successful': {'b': 2, 'd': 4}, 'Failed': {}}} -> {'OK': True, 'Value': 2}
      {'OK': True, 'Value': {'Successful': {}, 'Failed': {}}} ->
          {'Message': 'returnSingleResult: Failed and Successful dictionaries are empty', 'OK': False}
    """
    # if S_ERROR was returned, we return it as well
    if not dictRes["OK"]:
        return dictRes
    # if there is a Failed, we return the first one in an S_ERROR
    if "Failed" in dictRes["Value"] and len(dictRes["Value"]["Failed"]):
        errorMessage = list(dictRes["Value"]["Failed"].values())[0]
        if isinstance(errorMessage, dict):
            if isReturnStructure(errorMessage):
                return errorMessage
            else:
                return S_ERROR(str(errorMessage))
        return S_ERROR(errorMessage)
    # if there is a Successful, we return the first one in an S_OK
    elif "Successful" in dictRes["Value"] and len(dictRes["Value"]["Successful"]):
        return S_OK(list(dictRes["Value"]["Successful"].values())[0])
    else:
        return S_ERROR("returnSingleResult: Failed and Successful dictionaries are empty")


class SErrorException(Exception):
    """Exception class for use with `convertToReturnValue`"""

    def __init__(self, result):
        """Create a new exception return value

        If `result` is a `S_ERROR` return it directly else convert it to an
        appropriate value using `S_ERROR(result)`.

        :param result: The error to propagate
        """
        if not isSError(result):
            result = S_ERROR(result)
        self.result = result


def returnValueOrRaise(result):
    """Unwrap an S_OK/S_ERROR response into a value or Exception

    This method assists with using exceptions in DIRAC code by raising
    :exc:`SErrorException` if `result` is an error. This can then by propagated
    automatically as an `S_ERROR` by wrapping public facing functions with
    `@convertToReturnValue`.

    :param result: Result of a DIRAC function which returns `S_OK`/`S_ERROR`
    :returns: The value associated with the `S_OK` object
    :raises: If `result["OK"]` is falsey the original exception is re-raised.
             If no exception is known an :exc:`SErrorException` is raised.
    """
    if not result["OK"]:
        if "ExecInfo" in result:
            six.reraise(*result["ExecInfo"])
        else:
            raise SErrorException(result)
    return result["Value"]


def convertToReturnValue(func):
    """Decorate a function to convert return values to `S_OK`/`S_ERROR`

    If `func` returns, wrap the return value in `S_OK`.
    If `func` raises :exc:`SErrorException`, return the associated `S_ERROR`
    If `func` raises any other exception type, convert it to an `S_ERROR` object

    :param result: The bare result of a function call
    :returns: `S_OK`/`S_ERROR`
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return S_OK(func(*args, **kwargs))
        except SErrorException as e:
            return e.result
        except Exception as e:
            retval = S_ERROR(repr(e))
            # Replace CallStack with the one from the exception
            exc_type, exc_value, exc_tb = sys.exc_info()
            retval["ExecInfo"] = exc_type, exc_value, exc_tb
            retval["CallStack"] = traceback.format_tb(exc_tb)
            return retval

    return wrapped



""" PilotAgentsDB class is a front-end to the Pilot Agent Database.
    This database keeps track of all the submitted grid pilot jobs.
    It also registers the mapping of the DIRAC jobs to the pilot
    agents.

    Available methods are:

    addPilotTQReference()
    setPilotStatus()
    deletePilot()
    clearPilots()
    setPilotDestinationSite()
    storePilotOutput()
    getPilotOutput()
    setJobForPilot()
    getPilotsSummary()
    getGroupedPilotSummary()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import threading
import decimal

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
import DIRAC.Core.Utilities.Time as Time
from DIRAC.Core.Utilities import DErrno
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUsernameForDN, getDNForUsername, getVOForGroup
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
from DIRAC.Core.Utilities.MySQL import _quotedList
from DIRAC.WorkloadManagementSystem.Client import PilotStatus


class PilotAgentsDB(DB):
    def __init__(self):

        super(PilotAgentsDB, self).__init__("PilotAgentsDB", "WorkloadManagement/PilotAgentsDB")
        self.lock = threading.Lock()

    ##########################################################################################
    def addPilotTQReference(
        self, pilotRef, taskQueueID, ownerDN, ownerGroup, broker="Unknown", gridType="DIRAC", pilotStampDict={}
    ):
        """Add a new pilot job reference"""

        err = "PilotAgentsDB.addPilotTQReference: Failed to retrieve a new Id."

        for ref in pilotRef:
            stamp = ""
            if ref in pilotStampDict:
                stamp = pilotStampDict[ref]

            res = self._escapeString(ownerDN)
            if not res["OK"]:
                return res
            escapedOwnerDN = res["Value"]

            req = (
                "INSERT INTO PilotAgents( PilotJobReference, TaskQueueID, OwnerDN, "
                + "OwnerGroup, Broker, GridType, SubmissionTime, LastUpdateTime, Status, PilotStamp ) "
                + "VALUES ('%s',%d,%s,'%s','%s','%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),'Submitted','%s')"
                % (ref, int(taskQueueID), escapedOwnerDN, ownerGroup, broker, gridType, stamp)
            )

            result = self._update(req)
            if not result["OK"]:
                return result

            if "lastRowId" not in result:
                return S_ERROR("%s" % err)

        return S_OK()

    ##########################################################################################
    def setPilotStatus(
        self,
        pilotRef,
        status,
        destination=None,
        statusReason=None,
        gridSite=None,
        queue=None,
        benchmark=None,
        currentJob=None,
        updateTime=None,
        conn=False,
    ):
        """Set pilot job status"""

        setList = []
        setList.append("Status='%s'" % status)
        if updateTime:
            setList.append("LastUpdateTime='%s'" % updateTime)
        else:
            setList.append("LastUpdateTime=UTC_TIMESTAMP()")
        if not statusReason:
            statusReason = "Not given"
        setList.append("StatusReason='%s'" % statusReason)
        if gridSite:
            setList.append("GridSite='%s'" % gridSite)
        if queue:
            setList.append("Queue='%s'" % queue)
        if benchmark:
            setList.append("BenchMark='%s'" % float(benchmark))
        if currentJob:
            setList.append("CurrentJobID='%s'" % int(currentJob))
        if destination:
            setList.append("DestinationSite='%s'" % destination)
            if not gridSite:
                res = getCESiteMapping(destination)
                if res["OK"] and res["Value"]:
                    setList.append("GridSite='%s'" % res["Value"][destination])

        set_string = ",".join(setList)
        req = "UPDATE PilotAgents SET " + set_string + " WHERE PilotJobReference='%s'" % pilotRef
        result = self._update(req, conn=conn)
        if not result["OK"]:
            return result

        return S_OK()

    # ###########################################################################################
    # FIXME: this can't work ATM because of how the DB table is made. Maybe it would be useful later.
    #   def setPilotStatusBulk(self, pilotRefsStatusDict=None, statusReason=None,
    #                          conn=False):
    #     """ Set pilot job status in a bulk
    #     """
    #     if not pilotRefsStatusDict:
    #       return S_OK()

    #     # Building the request with "ON DUPLICATE KEY UPDATE"
    #     reqBase = "INSERT INTO PilotAgents (PilotJobReference, Status, StatusReason) VALUES "

    #     for pilotJobReference, status in pilotRefsStatusDict.items():
    #       req = reqBase + ','.join("('%s', '%s', '%s')" % (pilotJobReference, status, statusReason))
    #       req += " ON DUPLICATE KEY UPDATE Status=VALUES(Status),StatusReason=VALUES(StatusReason)"

    #     return self._update(req, conn)

    ##########################################################################################
    def selectPilots(
        self, condDict, older=None, newer=None, timeStamp="SubmissionTime", orderAttribute=None, limit=None
    ):
        """Select pilot references according to the provided criteria. "newer" and "older"
        specify the time interval in minutes
        """

        condition = self.buildCondition(condDict, older, newer, timeStamp)
        if orderAttribute:
            orderType = None
            orderField = orderAttribute
            if orderAttribute.find(":") != -1:
                orderType = orderAttribute.split(":")[1].upper()
                orderField = orderAttribute.split(":")[0]
            condition = condition + " ORDER BY " + orderField
            if orderType:
                condition = condition + " " + orderType

        if limit:
            condition = condition + " LIMIT " + str(limit)

        req = "SELECT PilotJobReference from PilotAgents"
        if condition:
            req += " %s " % condition
        result = self._query(req)
        if not result["OK"]:
            return result

        pilotList = []
        if result["Value"]:
            pilotList = [x[0] for x in result["Value"]]

        return S_OK(pilotList)

    ##########################################################################################
    def countPilots(self, condDict, older=None, newer=None, timeStamp="SubmissionTime"):
        """Select pilot references according to the provided criteria. "newer" and "older"
        specify the time interval in minutes
        """

        condition = self.buildCondition(condDict, older, newer, timeStamp)

        req = "SELECT COUNT(PilotID) from PilotAgents"
        if condition:
            req += " %s " % condition
        result = self._query(req)
        if not result["OK"]:
            return result

        return S_OK(result["Value"][0][0])

    #########################################################################################
    def getPilotGroups(self, groupList=["Status", "OwnerDN", "OwnerGroup", "GridType"], condDict={}):
        """
        Get all exisiting combinations of groupList Values
        """

        cmd = "SELECT %s from PilotAgents " % ", ".join(groupList)

        condList = []
        for cond in condDict:
            condList.append('%s in ( "%s" )' % (cond, '", "'.join([str(y) for y in condDict[cond]])))

        # the conditions should be escaped before hand, so it is not really nice to expose it this way...
        if condList:
            cmd += " WHERE %s " % " AND ".join(condList)

        cmd += " GROUP BY %s" % ", ".join(groupList)

        return self._query(cmd)

    ##########################################################################################
    def deletePilots(self, pilotIDs, conn=False):
        """Delete Pilots with IDs in the given list from the PilotAgentsDB"""

        if not isinstance(pilotIDs, list):
            return S_ERROR("Input argument is not a List")

        failed = []

        result = self._escapeValues(pilotIDs)
        if not result["OK"]:
            return S_ERROR("Failed to remove pilot: %s" % result["Value"])
        stringIDs = ",".join(result["Value"])
        for table in ["PilotOutput", "JobToPilotMapping", "PilotAgents"]:
            result = self._update("DELETE FROM %s WHERE PilotID in (%s)" % (table, stringIDs), conn=conn)
            if not result["OK"]:
                failed.append(table)

        if failed:
            return S_ERROR("Failed to remove pilot from %s tables" % ", ".join(failed))
        return S_OK(pilotIDs)

    ##########################################################################################
    def deletePilot(self, pilotRef, conn=False):
        """Delete Pilot with the given reference from the PilotAgentsDB"""

        if isinstance(pilotRef, six.string_types):
            pilotID = self.__getPilotID(pilotRef)
        else:
            pilotID = pilotRef

        return self.deletePilots([pilotID], conn=conn)

    ##########################################################################################
    def clearPilots(self, interval=30, aborted_interval=7):
        """Delete all the pilot references submitted before <interval> days"""

        reqList = []
        reqList.append(
            "SELECT PilotID FROM PilotAgents WHERE SubmissionTime < DATE_SUB(UTC_TIMESTAMP(),INTERVAL %d DAY)"
            % interval
        )
        reqList.append(
            "SELECT PilotID FROM PilotAgents WHERE Status='Aborted' \
AND SubmissionTime < DATE_SUB(UTC_TIMESTAMP(),INTERVAL %d DAY)"
            % aborted_interval
        )

        idList = None

        for req in reqList:
            result = self._query(req)
            if not result["OK"]:
                self.log.warn("Error while clearing up pilots")
            else:
                if result["Value"]:
                    idList = [x[0] for x in result["Value"]]
                    result = self.deletePilots(idList)
                    if not result["OK"]:
                        self.log.warn("Error while deleting pilots")

        return S_OK(idList)

    ##########################################################################################
    def getPilotInfo(self, pilotRef=False, parentId=False, conn=False, paramNames=[], pilotID=False):
        """Get all the information for the pilot job reference or reference list"""

        parameters = (
            [
                "PilotJobReference",
                "OwnerDN",
                "OwnerGroup",
                "GridType",
                "Broker",
                "Status",
                "DestinationSite",
                "BenchMark",
                "ParentID",
                "OutputReady",
                "AccountingSent",
                "SubmissionTime",
                "PilotID",
                "LastUpdateTime",
                "TaskQueueID",
                "GridSite",
                "PilotStamp",
                "Queue",
            ]
            if not paramNames
            else paramNames
        )

        cmd = "SELECT %s FROM PilotAgents" % ", ".join(parameters)
        condSQL = []
        for key, value in [("PilotJobReference", pilotRef), ("PilotID", pilotID), ("ParentID", parentId)]:
            resList = []
            for v in value if isinstance(value, list) else [value] if value else []:
                result = self._escapeString(v)
                if not result["OK"]:
                    return result
                resList.append(result["Value"])
            if resList:
                condSQL.append("%s IN (%s)" % (key, ",".join(resList)))
        if condSQL:
            cmd = "%s WHERE %s" % (cmd, " AND ".join(condSQL))

        result = self._query(cmd, conn=conn)
        if not result["OK"]:
            return result
        if not result["Value"]:
            msg = "No pilots found"
            if pilotRef:
                msg += " for PilotJobReference(s): %s" % pilotRef
            if parentId:
                msg += " with parent id: %s" % parentId
            return S_ERROR(DErrno.EWMSNOPILOT, msg)

        resDict = {}
        pilotIDs = []
        for row in result["Value"]:
            pilotDict = {}
            for i in range(len(parameters)):
                pilotDict[parameters[i]] = row[i]
                if parameters[i] == "PilotID":
                    pilotIDs.append(row[i])
            resDict[row[0]] = pilotDict

        result = self.getJobsForPilot(pilotIDs)
        if not result["OK"]:
            return S_OK(resDict)

        jobsDict = result["Value"]
        for pilotRef in resDict:
            pilotInfo = resDict[pilotRef]
            pilotID = pilotInfo["PilotID"]
            if pilotID in jobsDict:
                pilotInfo["Jobs"] = jobsDict[pilotID]

        return S_OK(resDict)

    ##########################################################################################
    def setPilotDestinationSite(self, pilotRef, destination, conn=False):
        """Set the pilot agent destination site"""

        gridSite = "Unknown"
        res = getCESiteMapping(destination)
        if res["OK"] and res["Value"]:
            gridSite = res["Value"][destination]

        req = "UPDATE PilotAgents SET DestinationSite='%s', GridSite='%s' WHERE PilotJobReference='%s'"
        req = req % (destination, gridSite, pilotRef)
        return self._update(req, conn=conn)

    ##########################################################################################
    def setPilotBenchmark(self, pilotRef, mark):
        """Set the pilot agent benchmark"""

        req = "UPDATE PilotAgents SET BenchMark='%f' WHERE PilotJobReference='%s'" % (mark, pilotRef)
        result = self._update(req)
        return result

    ##########################################################################################
    def setAccountingFlag(self, pilotRef, mark="True"):
        """Set the pilot AccountingSent flag"""

        req = "UPDATE PilotAgents SET AccountingSent='%s' WHERE PilotJobReference='%s'" % (mark, pilotRef)
        result = self._update(req)
        return result

    ##########################################################################################
    def storePilotOutput(self, pilotRef, output, error):
        """Store standard output and error for a pilot with pilotRef"""
        pilotID = self.__getPilotID(pilotRef)
        if not pilotID:
            return S_ERROR("Pilot reference not found %s" % pilotRef)

        result = self._escapeString(output)
        if not result["OK"]:
            return S_ERROR("Failed to escape output string")
        e_output = result["Value"]
        result = self._escapeString(error)
        if not result["OK"]:
            return S_ERROR("Failed to escape error string")
        e_error = result["Value"]
        req = "INSERT INTO PilotOutput (PilotID,StdOutput,StdError) VALUES (%d,%s,%s)" % (pilotID, e_output, e_error)
        result = self._update(req)
        if not result["OK"]:
            return result
        req = "UPDATE PilotAgents SET OutputReady='True' where PilotID=%d" % pilotID
        return self._update(req)

    ##########################################################################################
    def getPilotOutput(self, pilotRef):
        """Retrieve standard output and error for pilot with pilotRef"""

        req = "SELECT StdOutput, StdError FROM PilotOutput,PilotAgents WHERE "
        req += "PilotOutput.PilotID = PilotAgents.PilotID AND PilotAgents.PilotJobReference='%s'" % pilotRef
        result = self._query(req)
        if not result["OK"]:
            return result
        else:
            if result["Value"]:
                stdout = result["Value"][0][0]
                error = result["Value"][0][1]
                if stdout == '""':
                    stdout = ""
                if error == '""':
                    error = ""
                return S_OK({"StdOut": stdout, "StdErr": error})
            else:
                return S_ERROR("PilotJobReference " + pilotRef + " not found")

    ##########################################################################################
    def __getPilotID(self, pilotRef):
        """Get Pilot ID for the given pilot reference or a list of references"""

        if isinstance(pilotRef, six.string_types):
            req = "SELECT PilotID from PilotAgents WHERE PilotJobReference='%s'" % pilotRef
            result = self._query(req)
            if not result["OK"]:
                return 0
            else:
                if result["Value"]:
                    return int(result["Value"][0][0])
                return 0
        else:
            refString = ",".join(["'" + ref + "'" for ref in pilotRef])
            req = "SELECT PilotID from PilotAgents WHERE PilotJobReference in ( %s )" % refString
            result = self._query(req)
            if not result["OK"]:
                return []
            if result["Value"]:
                return [x[0] for x in result["Value"]]
            return []

    ##########################################################################################
    def setJobForPilot(self, jobID, pilotRef, site=None, updateStatus=True):
        """Store the jobID of the job executed by the pilot with reference pilotRef"""

        pilotID = self.__getPilotID(pilotRef)
        if pilotID:
            if updateStatus:
                reason = "Report from job %d" % int(jobID)
                result = self.setPilotStatus(pilotRef, status=PilotStatus.RUNNING, statusReason=reason, gridSite=site)
                if not result["OK"]:
                    return result
            req = "INSERT INTO JobToPilotMapping (PilotID,JobID,StartTime) VALUES (%d,%d,UTC_TIMESTAMP())" % (
                pilotID,
                jobID,
            )
            return self._update(req)
        else:
            return S_ERROR("PilotJobReference " + pilotRef + " not found")

    ##########################################################################################
    def setCurrentJobID(self, pilotRef, jobID):
        """Set the pilot agent current DIRAC job ID"""

        req = "UPDATE PilotAgents SET CurrentJobID=%d WHERE PilotJobReference='%s'" % (jobID, pilotRef)
        return self._update(req)

    ##########################################################################################
    def getJobsForPilot(self, pilotID):
        """Get IDs of Jobs that were executed by a pilot"""
        cmd = "SELECT pilotID,JobID FROM JobToPilotMapping "
        if isinstance(pilotID, list):
            cmd = cmd + " WHERE pilotID IN (%s)" % ",".join(["%s" % x for x in pilotID])
        else:
            cmd = cmd + " WHERE pilotID = %s" % pilotID

        result = self._query(cmd)
        if not result["OK"]:
            return result

        resDict = {}
        for row in result["Value"]:
            if not row[0] in resDict:
                resDict[row[0]] = []
            resDict[row[0]].append(row[1])
        return S_OK(resDict)

    ##########################################################################################
    def getPilotsForTaskQueue(self, taskQueueID, gridType=None, limit=None):
        """Get IDs of Pilot Agents that were submitted for the given taskQueue,
        specify optionally the grid type, results are sorted by Submission time
        an Optional limit can be set.
        """

        if gridType:
            req = "SELECT PilotID FROM PilotAgents WHERE TaskQueueID=%s AND GridType='%s' " % (taskQueueID, gridType)
        else:
            req = "SELECT PilotID FROM PilotAgents WHERE TaskQueueID=%s " % taskQueueID

        req += "ORDER BY SubmissionTime DESC "

        if limit:
            req += "LIMIT %s" % limit

        result = self._query(req)
        if not result["OK"]:
            return result
        if result["Value"]:
            pilotList = [x[0] for x in result["Value"]]
            return S_OK(pilotList)
        return S_ERROR("PilotJobReferences for TaskQueueID %s not found" % taskQueueID)

    ##########################################################################################
    def getPilotsForJobID(self, jobID):
        """Get ID of Pilot Agent that is running a given JobID"""

        result = self._query("SELECT PilotID FROM JobToPilotMapping WHERE JobID=%s" % jobID)

        if not result["OK"]:
            self.log.error("getPilotsForJobID failed", result["Message"])
            return result

        if result["Value"]:
            pilotList = [x[0] for x in result["Value"]]
            return S_OK(pilotList)
        self.log.verbose("PilotID for job not found: either not matched yet, or already deleted", "id=%s" % jobID)
        return S_OK([])

    ##########################################################################################
    def getPilotCurrentJob(self, pilotRef):
        """The job ID currently executed by the pilot"""
        req = "SELECT CurrentJobID FROM PilotAgents WHERE PilotJobReference='%s' " % pilotRef

        result = self._query(req)
        if not result["OK"]:
            return result
        if result["Value"]:
            jobID = int(result["Value"][0][0])
            return S_OK(jobID)
        self.log.warn("Current job ID for pilot %s is not known: pilot did not match jobs yet?" % pilotRef)
        return S_OK()

    ##########################################################################################
    # FIXME: investigate it getPilotSummaryShort can replace this method
    def getPilotSummary(self, startdate="", enddate=""):
        """Get summary of the pilot jobs status by site"""
        summary_dict = {}
        summary_dict["Total"] = {}

        for st in PilotStatus.PILOT_STATES:
            summary_dict["Total"][st] = 0
            req = "SELECT DestinationSite,count(DestinationSite) FROM PilotAgents " + "WHERE Status='%s' " % st
            if startdate:
                req = req + " AND SubmissionTime >= '%s'" % startdate
            if enddate:
                req = req + " AND SubmissionTime <= '%s'" % enddate

            req = req + " GROUP BY DestinationSite"
            result = self._query(req)
            if not result["OK"]:
                return result
            else:
                if result["Value"]:
                    for res in result["Value"]:
                        site = res[0]
                        count = res[1]
                        if site:
                            if site not in summary_dict:
                                summary_dict[site] = {}
                            summary_dict[site][st] = int(count)
                            summary_dict["Total"][st] += int(count)

        # Get aborted pilots in the last hour, day
        req = "SELECT DestinationSite,count(DestinationSite) FROM PilotAgents WHERE Status='Aborted' AND "
        reqDict = {}
        reqDict["Aborted_Hour"] = req + " LastUpdateTime >= DATE_SUB(UTC_TIMESTAMP(), INTERVAL 1 HOUR)"
        reqDict["Aborted_Day"] = req + " LastUpdateTime >= DATE_SUB(UTC_TIMESTAMP(), INTERVAL 1 DAY)"

        for key, req in reqDict.items():
            result = self._query(req)
            if not result["OK"]:
                break
            if result["Value"]:
                for res in result["Value"]:
                    site = res[0]
                    count = res[1]
                    if site:
                        if site in summary_dict:
                            summary_dict[site][key] = int(count)

        return S_OK(summary_dict)

    #   def getPilotSummaryShort( self, startTimeWindow = None, endTimeWindow = None, ce = '' ):
    #     """
    #     Spin off the method getPilotSummary. It is doing things in such a way that
    #     do not make much sense. This method returns the pilots that were updated in the
    #     time window [ startTimeWindow, endTimeWindow ), if they are present.
    #     """
    #
    #     sqlSelect = 'SELECT DestinationSite,Status,count(Status) FROM PilotAgents'
    #
    #     whereSelect = []
    #
    #     if startTimeWindow is not None:
    #       whereSelect.append( ' LastUpdateTime >= "%s"' % startTimeWindow )
    #     if endTimeWindow is not None:
    #       whereSelect.append( ' LastUpdateTime < "%s"' % endTimeWindow )
    #     if ce:
    #       whereSelect.append( ' DestinationSite = "%s"' % ce )
    #
    #     if whereSelect:
    #       sqlSelect += ' WHERE'
    #       sqlSelect += ' AND'.join( whereSelect )
    #
    #     sqlSelect += ' GROUP BY DestinationSite,Status'
    #
    #     resSelect = self._query( sqlSelect )
    #     if not resSelect[ 'OK' ]:
    #       return resSelect
    #
    #     result = { 'Total' : collections.defaultdict( int ) }
    #
    #     for row in resSelect[ 'Value' ]:
    #
    #       ceName, statusName, statusCount = row
    #
    #       if not ceName in result:
    #         result[ ceName ] = {}
    #       result[ ceName ][ statusName ] = int( statusCount )
    #
    #       result[ 'Total' ][ statusName ] += int( statusCount )
    #
    #     return S_OK( result )

    ##########################################################################################
    def getGroupedPilotSummary(self, selectDict, columnList):
        """
        The simplified pilot summary based on getPilotSummaryWeb method. It calculates pilot efficiency
        based on the same algorithm as in the Web version, basically takes into account Done and
        Aborted pilots only from the last day. The selection is done entirely in SQL.

        :param dict selectDict: A dictionary to pass additional conditions to select statements, i.e.
                                it allows to define start time for Done and Aborted Pilots. Unused.
        :param list columnList: A list of column to consider when grouping to calculate efficiencies.
                           e.g. ['GridSite', 'DestinationSite'] is used to calculate efficiencies
                           for sites and  CEs. If we want to add an OwnerGroup it would be:
                           ['GridSite', 'DestinationSite', 'OwnerGroup'].
        :return: S_OK/S_ERROR with a dict containing the ParameterNames and Records lists.
        """

        # TODO:
        #  add startItem and maxItems to the argument list
        #  limit output to  finalDict['Records'] = records[startItem:startItem + maxItems]
        table = PivotedPilotSummaryTable(columnList)
        sqlQuery = table.buildSQL()

        self.log.debug("SQL query : ")
        self.log.debug("\n" + sqlQuery)
        res = self._query(sqlQuery)
        if not res["OK"]:
            return res

        rows = []
        columns = table.getColumnList()
        try:
            groupIndex = columns.index("OwnerGroup")
            # should probably change a column name to VO here as well to avoid confusion
        except ValueError:
            groupIndex = None
        result = {"ParameterNames": columns}
        multiple = False
        # If not grouped by CE:
        if "CE" not in columns:
            multiple = True

        for row in res["Value"]:
            lrow = list(row)
            if groupIndex:
                lrow[groupIndex] = getVOForGroup(row[groupIndex])
            if multiple:
                lrow.append("Multiple")
            for index, value in enumerate(row):
                if isinstance(value, decimal.Decimal):
                    lrow[index] = float(value)
            # get the value of the Total column
            if "Total" in columnList:
                total = lrow[columnList.index("Total")]
            else:
                total = 0
            if "PilotJobEff" in columnList:
                eff = lrow[columnList.index("PilotJobEff")]
            else:
                eff = 0.0
            lrow.append(self._getElementStatus(total, eff))
            rows.append(list(lrow))
        # If not grouped by CE and more then 1 CE in the result:
        if multiple:
            columns.append("CE")  # 'DestinationSite' re-mapped to 'CE' already
        columns.append("Status")
        result["Records"] = rows
        result["TotalRecords"] = len(rows)
        return S_OK(result)

    def _getElementStatus(self, total, eff):
        """
        Assign status to a site or resource based on pilot efficiency.
        :param total: number of pilots to assign the status, otherwise 'Idle'
        :param eff:  efficiency in %

        :return: status string
        """

        # Evaluate the quality status of the Site/CE
        if total > 10:
            if eff < 25.0:
                return "Bad"
            elif eff < 60.0:
                return "Poor"
            elif eff < 85.0:
                return "Fair"
            else:
                return "Good"
        else:
            return "Idle"

    def getPilotSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        """Get summary of the pilot jobs status by CE/site in a standard structure"""
        allStateNames = PilotStatus.PILOT_STATES + ["Done_Empty", "Aborted_Hour"]
        paramNames = ["Site", "CE"] + allStateNames

        last_update = None
        if "LastUpdateTime" in selectDict:
            last_update = selectDict["LastUpdateTime"]
            del selectDict["LastUpdateTime"]
        site_select = []
        if "GridSite" in selectDict:
            site_select = selectDict["GridSite"]
            if not isinstance(site_select, list):
                site_select = [site_select]
            del selectDict["GridSite"]

        status_select = []
        if "Status" in selectDict:
            status_select = selectDict["Status"]
            if not isinstance(status_select, list):
                status_select = [status_select]
            del selectDict["Status"]

        expand_site = ""
        if "ExpandSite" in selectDict:
            expand_site = selectDict["ExpandSite"]
            site_select = [expand_site]
            del selectDict["ExpandSite"]

        # Get all the data from the database with various selections
        result = self.getCounters(
            "PilotAgents",
            ["GridSite", "DestinationSite", "Status"],
            selectDict,
            newer=last_update,
            timeStamp="LastUpdateTime",
        )
        if not result["OK"]:
            return result

        last_update = Time.dateTime() - Time.hour
        selectDict["Status"] = PilotStatus.ABORTED
        resultHour = self.getCounters(
            "PilotAgents",
            ["GridSite", "DestinationSite", "Status"],
            selectDict,
            newer=last_update,
            timeStamp="LastUpdateTime",
        )
        if not resultHour["OK"]:
            return resultHour

        last_update = Time.dateTime() - Time.day
        selectDict["Status"] = [PilotStatus.ABORTED, PilotStatus.DONE]
        resultDay = self.getCounters(
            "PilotAgents",
            ["GridSite", "DestinationSite", "Status"],
            selectDict,
            newer=last_update,
            timeStamp="LastUpdateTime",
        )
        if not resultDay["OK"]:
            return resultDay
        selectDict["CurrentJobID"] = 0
        selectDict["Status"] = PilotStatus.DONE
        resultDayEmpty = self.getCounters(
            "PilotAgents",
            ["GridSite", "DestinationSite", "Status"],
            selectDict,
            newer=last_update,
            timeStamp="LastUpdateTime",
        )
        if not resultDayEmpty["OK"]:
            return resultDayEmpty

        ceMap = {}
        resMap = getCESiteMapping()
        if resMap["OK"]:
            ceMap = resMap["Value"]

        # Sort out different counters
        resultDict = {}
        resultDict["Unknown"] = {}
        for attDict, count in result["Value"]:
            site = attDict["GridSite"]
            ce = attDict["DestinationSite"]
            state = attDict["Status"]
            if site == "Unknown" and ce != "Unknown" and ce != "Multiple" and ce in ceMap:
                site = ceMap[ce]
            if site not in resultDict:
                resultDict[site] = {}
            if ce not in resultDict[site]:
                resultDict[site][ce] = {}
                for p in allStateNames:
                    resultDict[site][ce][p] = 0

            resultDict[site][ce][state] = count

        for attDict, count in resultDay["Value"]:
            site = attDict["GridSite"]
            ce = attDict["DestinationSite"]
            state = attDict["Status"]
            if site == "Unknown" and ce != "Unknown" and ce in ceMap:
                site = ceMap[ce]
            if state == PilotStatus.DONE:
                resultDict[site][ce][PilotStatus.DONE] = count
            if state == PilotStatus.ABORTED:
                resultDict[site][ce][PilotStatus.ABORTED] = count

        for attDict, count in resultDayEmpty["Value"]:
            site = attDict["GridSite"]
            ce = attDict["DestinationSite"]
            state = attDict["Status"]
            if site == "Unknown" and ce != "Unknown" and ce in ceMap:
                site = ceMap[ce]
            if state == PilotStatus.DONE:
                resultDict[site][ce]["Done_Empty"] = count

        for attDict, count in resultHour["Value"]:
            site = attDict["GridSite"]
            ce = attDict["DestinationSite"]
            state = attDict["Status"]
            if site == "Unknown" and ce != "Unknown" and ce in ceMap:
                site = ceMap[ce]
            if state == PilotStatus.ABORTED:
                resultDict[site][ce]["Aborted_Hour"] = count

        records = []
        siteSumDict = {}
        for site in resultDict:
            sumDict = {}
            for state in allStateNames:
                if state not in sumDict:
                    sumDict[state] = 0
            sumDict["Total"] = 0
            for ce in resultDict[site]:
                itemList = [site, ce]
                total = 0
                for state in allStateNames:
                    itemList.append(resultDict[site][ce][state])
                    sumDict[state] += resultDict[site][ce][state]
                    if state == PilotStatus.DONE:
                        done = resultDict[site][ce][state]
                    if state == "Done_Empty":
                        empty = resultDict[site][ce][state]
                    if state == PilotStatus.ABORTED:
                        aborted = resultDict[site][ce][state]
                    if state != "Aborted_Hour" and state != "Done_Empty":
                        total += resultDict[site][ce][state]

                sumDict["Total"] += total
                # Add the total number of pilots seen in the last day
                itemList.append(total)
                # Add pilot submission efficiency evaluation
                if (done - empty) > 0:
                    eff = done / (done - empty)
                elif done == 0:
                    eff = 0.0
                elif empty == done:
                    eff = 99.0
                else:
                    eff = 0.0
                itemList.append("%.2f" % eff)
                # Add pilot job efficiency evaluation
                if total > 0:
                    eff = (total - aborted) / total * 100
                else:
                    eff = 100.0
                itemList.append("%.2f" % eff)

                # Evaluate the quality status of the CE
                if total > 10:
                    if eff < 25.0:
                        itemList.append("Bad")
                    elif eff < 60.0:
                        itemList.append("Poor")
                    elif eff < 85.0:
                        itemList.append("Fair")
                    else:
                        itemList.append("Good")
                else:
                    itemList.append("Idle")

                if len(resultDict[site]) == 1 or expand_site:
                    records.append(itemList)

            if len(resultDict[site]) > 1 and not expand_site:
                itemList = [site, "Multiple"]
                for state in allStateNames + ["Total"]:
                    if state in sumDict:
                        itemList.append(sumDict[state])
                    else:
                        itemList.append(0)
                done = sumDict[PilotStatus.DONE]
                empty = sumDict["Done_Empty"]
                aborted = sumDict[PilotStatus.ABORTED]
                total = sumDict["Total"]

                # Add pilot submission efficiency evaluation
                if (done - empty) > 0:
                    eff = done / (done - empty)
                elif done == 0:
                    eff = 0.0
                elif empty == done:
                    eff = 99.0
                else:
                    eff = 0.0
                itemList.append("%.2f" % eff)
                # Add pilot job efficiency evaluation
                if total > 0:
                    eff = (total - aborted) / total * 100
                else:
                    eff = 100.0
                itemList.append("%.2f" % eff)

                # Evaluate the quality status of the Site
                if total > 10:
                    if eff < 25.0:
                        itemList.append("Bad")
                    elif eff < 60.0:
                        itemList.append("Poor")
                    elif eff < 85.0:
                        itemList.append("Fair")
                    else:
                        itemList.append("Good")
                else:
                    itemList.append("Idle")
                records.append(itemList)

            for state in allStateNames + ["Total"]:
                if state not in siteSumDict:
                    siteSumDict[state] = sumDict[state]
                else:
                    siteSumDict[state] += sumDict[state]

        # Perform site selection
        if site_select:
            new_records = []
            for r in records:
                if r[0] in site_select:
                    new_records.append(r)
            records = new_records

        # Perform status selection
        if status_select:
            new_records = []
            for r in records:
                if r[14] in status_select:
                    new_records.append(r)
            records = new_records

        # Get the Site Mask data
        result = SiteStatus().getUsableSites()
        if result["OK"]:
            siteMask = result["Value"]
            for r in records:
                if r[0] in siteMask:
                    r.append("Yes")
                else:
                    r.append("No")
        else:
            for r in records:
                r.append("Unknown")

        finalDict = {}
        finalDict["TotalRecords"] = len(records)
        finalDict["ParameterNames"] = paramNames + ["Total", "PilotsPerJob", "PilotJobEff", "Status", "InMask"]

        # Return all the records if maxItems == 0 or the specified number otherwise
        if maxItems:
            finalDict["Records"] = records[startItem : startItem + maxItems]
        else:
            finalDict["Records"] = records

        done = siteSumDict[PilotStatus.DONE]
        empty = siteSumDict["Done_Empty"]
        aborted = siteSumDict[PilotStatus.ABORTED]
        total = siteSumDict["Total"]

        # Add pilot submission efficiency evaluation
        if (done - empty) > 0:
            eff = done / (done - empty)
        elif done == 0:
            eff = 0.0
        elif empty == done:
            eff = 99.0
        else:
            eff = 0.0
        siteSumDict["PilotsPerJob"] = "%.2f" % eff
        # Add pilot job efficiency evaluation
        if total > 0:
            eff = (total - aborted) / total * 100
        else:
            eff = 100.0
        siteSumDict["PilotJobEff"] = "%.2f" % eff

        # Evaluate the overall quality status
        if total > 100:
            if eff < 25.0:
                siteSumDict["Status"] = "Bad"
            elif eff < 60.0:
                siteSumDict["Status"] = "Poor"
            elif eff < 85.0:
                siteSumDict["Status"] = "Fair"
            else:
                siteSumDict["Status"] = "Good"
        else:
            siteSumDict["Status"] = "Idle"
        finalDict["Extras"] = siteSumDict

        return S_OK(finalDict)

    ##########################################################################################
    def getPilotMonitorSelectors(self):
        """Get distinct values for the Pilot Monitor page selectors"""

        paramNames = ["OwnerDN", "OwnerGroup", "GridType", "Broker", "Status", "DestinationSite", "GridSite"]

        resultDict = {}
        for param in paramNames:
            result = self.getDistinctAttributeValues("PilotAgents", param)
            if result["OK"]:
                resultDict[param] = result["Value"]
            else:
                resultDict = []

            if param == "OwnerDN":
                userList = []
                for dn in result["Value"]:
                    resultUser = getUsernameForDN(dn)
                    if resultUser["OK"]:
                        userList.append(resultUser["Value"])
                resultDict["Owner"] = userList

        return S_OK(resultDict)

    ##########################################################################################
    def getPilotMonitorWeb(self, selectDict, sortList, startItem, maxItems):
        """Get summary of the pilot job information in a standard structure"""

        resultDict = {}
        if "LastUpdateTime" in selectDict:
            del selectDict["LastUpdateTime"]
        if "Owner" in selectDict:
            userList = selectDict["Owner"]
            if not isinstance(userList, list):
                userList = [userList]
            dnList = []
            for uName in userList:
                uList = getDNForUsername(uName)["Value"]
                dnList += uList
            selectDict["OwnerDN"] = dnList
            del selectDict["Owner"]
        startDate = selectDict.get("FromDate", None)
        if startDate:
            del selectDict["FromDate"]
        # For backward compatibility
        if startDate is None:
            startDate = selectDict.get("LastUpdateTime", None)
            if startDate:
                del selectDict["LastUpdateTime"]
        endDate = selectDict.get("ToDate", None)
        if endDate:
            del selectDict["ToDate"]

        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = sortList[0][0] + ":" + sortList[0][1]
        else:
            orderAttribute = None

        # Select pilots for the summary
        result = self.selectPilots(
            selectDict, orderAttribute=orderAttribute, newer=startDate, older=endDate, timeStamp="LastUpdateTime"
        )
        if not result["OK"]:
            return S_ERROR("Failed to select pilots: " + result["Message"])

        pList = result["Value"]
        nPilots = len(pList)
        resultDict["TotalRecords"] = nPilots
        if nPilots == 0:
            return S_OK(resultDict)

        ini = startItem
        last = ini + maxItems
        if ini >= nPilots:
            return S_ERROR("Item number out of range")
        if last > nPilots:
            last = nPilots
        pilotList = pList[ini:last]

        paramNames = [
            "PilotJobReference",
            "OwnerDN",
            "OwnerGroup",
            "GridType",
            "Broker",
            "Status",
            "DestinationSite",
            "BenchMark",
            "ParentID",
            "SubmissionTime",
            "PilotID",
            "LastUpdateTime",
            "CurrentJobID",
            "TaskQueueID",
            "GridSite",
        ]

        result = self.getPilotInfo(pilotList, paramNames=paramNames)
        if not result["OK"]:
            return S_ERROR("Failed to get pilot info: " + result["Message"])

        pilotDict = result["Value"]
        records = []
        for pilot in pilotList:
            parList = []
            for parameter in paramNames:
                if not isinstance(pilotDict[pilot][parameter], six.integer_types):
                    parList.append(str(pilotDict[pilot][parameter]))
                else:
                    parList.append(pilotDict[pilot][parameter])
                if parameter == "GridSite":
                    gridSite = pilotDict[pilot][parameter]

            # If the Grid Site is unknown try to recover it in the last moment
            if gridSite == "Unknown":
                ce = pilotDict[pilot]["DestinationSite"]
                result = getCESiteMapping(ce)
                if result["OK"]:
                    gridSite = result["Value"].get(ce)
                    del parList[-1]
                    parList.append(gridSite)
            records.append(parList)

        resultDict["ParameterNames"] = paramNames
        resultDict["Records"] = records

        return S_OK(resultDict)


class PivotedPilotSummaryTable:
    """
    The class creates a 'pivoted' table by combining records with the same group
    of self.columnList into a single row. It allows an easy calculation of pilot efficiencies.
    """

    pstates = ["Submitted", "Done", "Failed", "Aborted", "Running", "Waiting", "Scheduled", "Ready"]

    def __init__(self, columnList):
        """
        Initialise a table with columns to be grouped by.

        :param columnList: i.e. ['GridSite', 'DestinationSite']
        :return:
        """

        self.columnList = columnList

        # we want 'Site' and 'CE' in the final result
        colMap = {"GridSite": "Site", "DestinationSite": "CE"}
        self._columns = [colMap.get(val, val) for val in columnList]

        self._columns += self.pstates  # MySQL._query() does not give us column names, sadly.

    def buildSQL(self, selectDict=None):
        """
        Build an SQL query to create a table with all status counts in one row, ("pivoted")
        grouped by columns in the column list.

        :param dict selectDict:
        :return: SQL query
        """

        lastUpdate = Time.dateTime() - Time.day
        lastHour = Time.dateTime() - Time.hour

        pvtable = "pivoted"
        innerGroupBy = (
            "(SELECT %s, Status,\n "
            "count(CASE WHEN CurrentJobID=0  THEN 1 END) AS Empties,"
            "count(CASE WHEN LastUpdateTime > '%s' THEN 1 END) AS Last_Hour,"
            " count(*) AS qty FROM PilotAgents\n "
            "WHERE Status NOT IN ('Done', 'Aborted') OR (Status in ('Done', 'Aborted') \n"
            " AND \n"
            " LastUpdateTime > '%s')"
            " GROUP by %s, Status)\n AS %s"
            % (_quotedList(self.columnList), lastHour, lastUpdate, _quotedList(self.columnList), pvtable)
        )

        # pivoted table: combine records with the same group of self.columnList into a single row.

        pivotedQuery = "SELECT %s,\n" % ", ".join([pvtable + "." + item for item in self.columnList])
        lineTemplate = " SUM(if (pivoted.Status={state!r}, pivoted.qty, 0)) AS {state}"
        pivotedQuery += ",\n".join(lineTemplate.format(state=state) for state in self.pstates)
        pivotedQuery += ",\n  SUM(if (pivoted.Status='Aborted', pivoted.Last_Hour, 0)) AS Aborted_Hour"
        pivotedQuery += (
            ",\n  SUM(if (%s.Status='Done', %s.Empties,0)) AS Done_Empty,\n"
            "  SUM(%s.qty) AS Total "
            "FROM\n" % (pvtable, pvtable, pvtable)
        )

        outerGroupBy = " GROUP BY %s) \nAS pivotedEff;" % _quotedList(self.columnList)

        # add efficiency columns using aliases defined in the pivoted table
        effCase = (
            "(CASE\n  WHEN pivotedEff.Done - pivotedEff.Done_Empty > 0 \n"
            "  THEN pivotedEff.Done/(pivotedEff.Done-pivotedEff.Done_Empty) \n"
            "  WHEN pivotedEff.Done=0 THEN 0 \n"
            "  WHEN pivotedEff.Done=pivotedEff.Done_Empty \n"
            "  THEN 99.0 ELSE 0.0 END) AS PilotsPerJob,\n"
            " (pivotedEff.Total - pivotedEff.Aborted)/pivotedEff.Total*100.0 AS PilotJobEff \nFROM \n("
        )
        effSelectTemplate = " CAST(pivotedEff.{state} AS UNSIGNED) AS {state} "
        # now select the columns + states:
        pivotedEff = (
            "SELECT %s,\n" % ", ".join(["pivotedEff" + "." + item for item in self.columnList])
            + ", ".join(effSelectTemplate.format(state=state) for state in self.pstates + ["Aborted_Hour", "Total"])
            + ", \n"
        )

        finalQuery = pivotedEff + effCase + pivotedQuery + innerGroupBy + outerGroupBy
        self._columns += ["Aborted_Hour", "Total", "PilotsPerJob", "PilotJobEff"]
        return finalQuery

    def getColumnList(self):

        return self._columns



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.AccountingSystem.private.Policies.JobPolicy import JobPolicy as myJobPolicy

gPoliciesList = {"Job": myJobPolicy(), "WMSHistory": myJobPolicy(), "Pilot": myJobPolicy(), "Null": False}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger


class FilterExecutor(object):

    ALLKW = "all"

    def __init__(self):
        self.__filters = {}
        self.__globalFilters = []

    def applyFilters(self, iD, credDict, condDict, groupingList):
        filters2Apply = list(self.__globalFilters)
        if iD in self.__filters:
            filters2Apply.extend(self.__filters[iD])
        for myFilter in filters2Apply:
            try:
                gLogger.info("Applying filter %s for %s" % (myFilter.__name__, iD))
                retVal = myFilter(credDict, condDict, groupingList)
                if not retVal["OK"]:
                    gLogger.info("Filter %s for %s failed: %s" % (myFilter.__name__, iD, retVal["Message"]))
                    return retVal
            except Exception:
                gLogger.exception("Exception while applying filter", "%s for %s" % (myFilter.__name__, iD))
                return S_ERROR("Exception while applying filters")
        return S_OK()

    def addFilter(self, iD, myFilter):
        if iD not in self.__filters:
            self.__filters[iD] = []
        if isinstance(myFilter, (list, tuple)):
            self.__filters[iD].extend(myFilter)
        else:
            self.__filters[iD].append(myFilter)

    def addGlobalFilter(self, myFilter):
        if isinstance(myFilter, (list, tuple)):
            self.__globalFilters.extend(myFilter)
        else:
            self.__globalFilters.append(myFilter)



""" The SGE TimeLeft utility interrogates the SGE batch system for the
    current CPU consumed, as well as its limit.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re
import time
import socket

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import runCommand
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage


class SGEResourceUsage(ResourceUsage):
    """
    This is the SGE plugin of the TimeLeft Utility
    """

    def __init__(self):
        """Standard constructor"""
        super(SGEResourceUsage, self).__init__("SGE", "JOB_ID")

        self.queue = os.environ.get("QUEUE")
        sgePath = os.environ.get("SGE_BINARY_PATH")
        if sgePath:
            os.environ["PATH"] += ":" + sgePath

        self.log.verbose("JOB_ID=%s, QUEUE=%s" % (self.jobID, self.queue))
        self.startTime = time.time()

    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries CPU, CPULimit,
        WallClock, WallClockLimit, and Unit for current slot.
        """
        cmd = "qstat -f -j %s" % (self.jobID)
        result = runCommand(cmd)
        if not result["OK"]:
            return result

        cpu = None
        cpuLimit = None
        wallClock = None
        wallClockLimit = None

        lines = str(result["Value"]).split("\n")
        for line in lines:
            if re.search("usage.*cpu.*", line):
                match = re.search(r"cpu=([\d,:]*),", line)
                if match:
                    cpuList = match.groups()[0].split(":")
                try:
                    newcpu = 0.0
                    if len(cpuList) == 3:
                        newcpu = float(cpuList[0]) * 3600 + float(cpuList[1]) * 60 + float(cpuList[2])
                    elif len(cpuList) == 4:
                        newcpu = (
                            float(cpuList[0]) * 24 * 3600
                            + float(cpuList[1]) * 3600
                            + float(cpuList[2]) * 60
                            + float(cpuList[3])
                        )
                    if not cpu or newcpu > cpu:
                        cpu = newcpu
                except ValueError:
                    self.log.warn('Problem parsing "%s" for CPU consumed' % line)
            if re.search("hard resource_list.*cpu.*", line):
                match = re.search(r"_cpu=(\d*)", line)
                if match:
                    cpuLimit = float(match.groups()[0])
                match = re.search(r"_rt=(\d*)", line)
                if match:
                    wallClockLimit = float(match.groups()[0])
            else:
                self.log.warn("No hard limits found")

        # Some SGE batch systems apply CPU scaling factor to the CPU consumption figures
        if cpu:
            factor = _getCPUScalingFactor()
            if factor:
                cpu = cpu / factor

        consumed = {"CPU": cpu, "CPULimit": cpuLimit, "WallClock": wallClock, "WallClockLimit": wallClockLimit}

        if None in consumed.values():
            missed = [key for key, val in consumed.items() if val is None]
            msg = "Could not determine parameter"
            self.log.warn("Could not determine parameter", ",".join(missed))
            self.log.debug("This is the stdout from the batch system call\n%s" % (result["Value"]))
        else:
            self.log.debug("TimeLeft counters complete:", str(consumed))

        if cpuLimit or wallClockLimit:
            # We have got a partial result from SGE
            if not cpuLimit:
                # Take some margin
                consumed["CPULimit"] = wallClockLimit * 0.8
            if not wallClockLimit:
                consumed["WallClockLimit"] = cpuLimit / 0.8
            if not cpu:
                consumed["CPU"] = time.time() - self.startTime
            if not wallClock:
                consumed["WallClock"] = time.time() - self.startTime
            self.log.debug("TimeLeft counters restored:", str(consumed))
            return S_OK(consumed)
        else:
            msg = "Could not determine necessary parameters"
            self.log.info(msg, ":\nThis is the stdout from the batch system call\n%s" % (result["Value"]))
            retVal = S_ERROR(msg)
            retVal["Value"] = consumed
            return retVal


def _getCPUScalingFactor():

    host = socket.getfqdn()
    cmd = "qconf -se %s" % host
    result = runCommand(cmd)
    if not result["OK"]:
        return None

    lines = str(result["Value"]).split("\n")
    for line in lines:
        if re.search("usage_scaling", line):
            match = re.search(r"cpu=([\d,\.]*),", line)
            if match:
                return float(match.groups()[0])
    return None



"""
This is the Data Integrity Client which allows the simple reporting of
problematic file and replicas to the IntegrityDB and their status
correctly updated in the FileCatalog.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Base.Client import Client, createClient

__RCSID__ = "$Id$"


@createClient("DataManagement/DataIntegrity")
class DataIntegrityClient(Client):
    """Client exposing the DataIntegrity Service."""

    def __init__(self, **kwargs):

        super(DataIntegrityClient, self).__init__(**kwargs)
        self.setServer("DataManagement/DataIntegrity")
        self.dm = DataManager()
        self.fc = FileCatalog()

    def setFileProblematic(self, lfn, reason, sourceComponent=""):
        """This method updates the status of the file in the FileCatalog and the IntegrityDB

        lfn - the lfn of the file
        reason - this is given to the integrity DB and should reflect the problem observed with the file

        sourceComponent is the component issuing the request.
        """
        if isinstance(lfn, list):
            lfns = lfn
        elif isinstance(lfn, six.string_types):
            lfns = [lfn]
        else:
            errStr = "DataIntegrityClient.setFileProblematic: Supplied file info must be list or a single LFN."
            gLogger.error(errStr)
            return S_ERROR(errStr)
        gLogger.info("DataIntegrityClient.setFileProblematic: Attempting to update %s files." % len(lfns))
        fileMetadata = {}
        for lfn in lfns:
            fileMetadata[lfn] = {"Prognosis": reason, "LFN": lfn, "PFN": "", "SE": ""}
        res = self.insertProblematic(sourceComponent, fileMetadata)
        if not res["OK"]:
            gLogger.error("DataIntegrityClient.setReplicaProblematic: Failed to insert problematics to integrity DB")
        return res

    def reportProblematicReplicas(self, replicaTuple, se, reason):
        """Simple wrapper function around setReplicaProblematic"""
        gLogger.info("The following %s files had %s at %s" % (len(replicaTuple), reason, se))
        for lfn, _pfn, se, reason in sorted(replicaTuple):
            if lfn:
                gLogger.info(lfn)
        res = self.setReplicaProblematic(replicaTuple, sourceComponent="DataIntegrityClient")
        if not res["OK"]:
            gLogger.info("Failed to update integrity DB with replicas", res["Message"])
        else:
            gLogger.info("Successfully updated integrity DB with replicas")

    def setReplicaProblematic(self, replicaTuple, sourceComponent=""):
        """This method updates the status of the replica in the FileCatalog and the IntegrityDB
        The supplied replicaDict should be of the form {lfn :{'PFN':pfn,'SE':se,'Prognosis':prognosis}

        lfn - the lfn of the file
        pfn - the pfn if available (otherwise '')
        se - the storage element of the problematic replica (otherwise '')
        prognosis - this is given to the integrity DB and should reflect the problem observed with the file

        sourceComponent is the component issuing the request.
        """
        if isinstance(replicaTuple, tuple):
            replicaTuple = [replicaTuple]
        elif isinstance(replicaTuple, list):
            pass
        else:
            errStr = (
                "DataIntegrityClient.setReplicaProblematic: Supplied replica info must be a tuple or list of tuples."
            )
            gLogger.error(errStr)
            return S_ERROR(errStr)
        gLogger.info("DataIntegrityClient.setReplicaProblematic: Attempting to update %s replicas." % len(replicaTuple))
        replicaDict = {}
        for lfn, pfn, se, reason in replicaTuple:
            replicaDict[lfn] = {"Prognosis": reason, "LFN": lfn, "PFN": pfn, "SE": se}
        res = self.insertProblematic(sourceComponent, replicaDict)
        if not res["OK"]:
            gLogger.error("DataIntegrityClient.setReplicaProblematic: Failed to insert problematic to integrity DB")
            return res
        for lfn in replicaDict.keys():
            replicaDict[lfn]["Status"] = "Problematic"

        res = self.fc.setReplicaStatus(replicaDict)
        if not res["OK"]:
            errStr = "DataIntegrityClient.setReplicaProblematic: Completely failed to update replicas."
            gLogger.error(errStr, res["Message"])
            return res
        failed = res["Value"]["Failed"]
        successful = res["Value"]["Successful"]
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    ##########################################################################
    #
    # This section contains the resolution methods for various prognoses
    #

    def __updateCompletedFiles(self, prognosis, fileID):
        gLogger.info("%s file (%d) is resolved" % (prognosis, fileID))
        return self.setProblematicStatus(fileID, "Resolved")

    def __returnProblematicError(self, fileID, res):
        self.incrementProblematicRetry(fileID)
        gLogger.error("DataIntegrityClient failure", res["Message"])
        return res

    def __updateReplicaToChecked(self, problematicDict):
        lfn = problematicDict["LFN"]
        fileID = problematicDict["FileID"]
        prognosis = problematicDict["Prognosis"]
        problematicDict["Status"] = "Checked"

        res = returnSingleResult(self.fc.setReplicaStatus({lfn: problematicDict}))

        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        gLogger.info("%s replica (%d) is updated to Checked status" % (prognosis, fileID))
        return self.__updateCompletedFiles(prognosis, fileID)

    def resolveCatalogPFNSizeMismatch(self, problematicDict):
        """This takes the problematic dictionary returned by the integrity DB and resolved the CatalogPFNSizeMismatch prognosis"""
        lfn = problematicDict["LFN"]
        se = problematicDict["SE"]
        fileID = problematicDict["FileID"]

        res = returnSingleResult(self.fc.getFileSize(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        catalogSize = res["Value"]
        res = returnSingleResult(StorageElement(se).getFileSize(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        storageSize = res["Value"]
        bkKCatalog = FileCatalog(["BookkeepingDB"])
        res = returnSingleResult(bkKCatalog.getFileSize(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        bookkeepingSize = res["Value"]
        if bookkeepingSize == catalogSize == storageSize:
            gLogger.info("CatalogPFNSizeMismatch replica (%d) matched all registered sizes." % fileID)
            return self.__updateReplicaToChecked(problematicDict)
        if catalogSize == bookkeepingSize:
            gLogger.info("CatalogPFNSizeMismatch replica (%d) found to mismatch the bookkeeping also" % fileID)
            res = returnSingleResult(self.fc.getReplicas(lfn))
            if not res["OK"]:
                return self.__returnProblematicError(fileID, res)
            if len(res["Value"]) <= 1:
                gLogger.info("CatalogPFNSizeMismatch replica (%d) has no other replicas." % fileID)
                return S_ERROR("Not removing catalog file mismatch since the only replica")
            else:
                gLogger.info("CatalogPFNSizeMismatch replica (%d) has other replicas. Removing..." % fileID)
                res = self.dm.removeReplica(se, lfn)
                if not res["OK"]:
                    return self.__returnProblematicError(fileID, res)
                return self.__updateCompletedFiles("CatalogPFNSizeMismatch", fileID)
        if (catalogSize != bookkeepingSize) and (bookkeepingSize == storageSize):
            gLogger.info("CatalogPFNSizeMismatch replica (%d) found to match the bookkeeping size" % fileID)
            res = self.__updateReplicaToChecked(problematicDict)
            if not res["OK"]:
                return self.__returnProblematicError(fileID, res)
            return self.changeProblematicPrognosis(fileID, "BKCatalogSizeMismatch")
        gLogger.info("CatalogPFNSizeMismatch replica (%d) all sizes found mismatch. Updating retry count" % fileID)
        return self.incrementProblematicRetry(fileID)

    ############################################################################################

    def _reportProblematicFiles(self, lfns, reason):
        """Simple wrapper function around setFileProblematic"""
        gLogger.info("The following %s files were found with %s" % (len(lfns), reason))
        for lfn in sorted(lfns):
            gLogger.info(lfn)
        res = self.setFileProblematic(lfns, reason, sourceComponent="DataIntegrityClient")
        if not res["OK"]:
            gLogger.info("Failed to update integrity DB with files", res["Message"])
        else:
            gLogger.info("Successfully updated integrity DB with files")



""" SQLAlchemyDB:
    This module provides the BaseRSSDB class for providing standard DB interactions.

    Uses sqlalchemy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import datetime
from sqlalchemy import create_engine, desc, exc
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.query import Query

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DIRACDB import DIRACDB
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters


class SQLAlchemyDB(DIRACDB):
    """
    Base class that defines some of the basic DB interactions.
    """

    def __init__(self, *args, **kwargs):
        """c'tor

        :param self: self reference
        """
        self.fullname = self.__class__.__name__
        super(SQLAlchemyDB, self).__init__(*args, **kwargs)

        self.extensions = gConfig.getValue("DIRAC/Extensions", [])
        self.tablesList = []

    def _initializeConnection(self, dbPath):
        """
        Collect from the CS all the info needed to connect to the DB.
        """

        result = getDBParameters(dbPath)
        if not result["OK"]:
            raise Exception("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.log.debug("db parameters: %s" % dbParameters)
        self.host = dbParameters["Host"]
        self.port = dbParameters["Port"]
        self.user = dbParameters["User"]
        self.password = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

        self.engine = create_engine(
            "mysql://%s:%s@%s:%s/%s" % (self.user, self.password, self.host, self.port, self.dbName),
            pool_recycle=3600,
            echo_pool=True,
            echo=self.log.getLevel() == "DEBUG",
        )
        self.sessionMaker_o = sessionmaker(bind=self.engine)
        self.inspector = Inspector.from_engine(self.engine)

    def _createTablesIfNotThere(self, tablesList):
        """
        Adds each table in tablesList to the DB if not already present
        """
        tablesInDB = self.inspector.get_table_names()

        for table in tablesList:
            if table not in tablesInDB:
                found = False
                # is it in the extension? (fully or extended)
                for ext in self.extensions:
                    try:
                        getattr(
                            __import__(ext + self.__class__.__module__, globals(), locals(), [table]), table
                        ).__table__.create(
                            self.engine
                        )  # pylint: disable=no-member
                        found = True
                        break
                    except (ImportError, AttributeError):
                        continue
                # If not found in extensions, import it from DIRAC base.
                if not found:
                    getattr(
                        __import__(self.__class__.__module__, globals(), locals(), [table]), table
                    ).__table__.create(
                        self.engine
                    )  # pylint: disable=no-member
            else:
                gLogger.debug("Table %s already exists" % table)

    def insert(self, table, params):
        """
        Inserts params in the DB.

        :param str table: table where to insert
        :param dict params: Dictionary to fill a single line

        :return: S_OK() || S_ERROR()
        """

        # expire_on_commit is set to False so that we can still use the object after we close the session
        session = self.sessionMaker_o(expire_on_commit=False)  # FIXME: should we use this flag elsewhere?

        found = False
        for ext in self.extensions:
            try:
                tableRow_o = getattr(__import__(ext + self.__class__.__module__, globals(), locals(), [table]), table)()
                found = True
                break
            except (ImportError, AttributeError):
                continue
        # If not found in extensions, import it from DIRAC base (this same module).
        if not found:
            tableRow_o = getattr(__import__(self.__class__.__module__, globals(), locals(), [table]), table)()

        tableRow_o.fromDict(params)

        try:
            session.add(tableRow_o)
            session.commit()
            return S_OK()
        except exc.IntegrityError as err:
            self.log.warn("insert: trying to insert a duplicate key? %s" % err)
            session.rollback()
        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("insert: unexpected exception", lException=e)
            return S_ERROR("insert: unexpected exception %s" % e)
        finally:
            session.close()

    def select(self, table, params):
        """
        Uses params to build conditional SQL statement ( WHERE ... ).

        :Parameters:
          **params** - `dict`
            arguments for the mysql query ( must match table columns ! ).

        :return: S_OK() || S_ERROR()
        """

        session = self.sessionMaker_o()

        # finding the table
        found = False
        for ext in self.extensions:
            try:
                table_c = getattr(__import__(ext + self.__class__.__module__, globals(), locals(), [table]), table)
                found = True
                break
            except (ImportError, AttributeError):
                continue
        # If not found in extensions, import it from DIRAC base (this same module).
        if not found:
            table_c = getattr(__import__(self.__class__.__module__, globals(), locals(), [table]), table)

        # handling query conditions found in 'Meta'
        columnNames = [column.lower() for column in params.get("Meta", {}).get("columns", [])]
        older = params.get("Meta", {}).get("older", None)
        newer = params.get("Meta", {}).get("newer", None)
        order = params.get("Meta", {}).get("order", None)
        limit = params.get("Meta", {}).get("limit", None)
        params.pop("Meta", None)

        try:
            # setting up the select query
            if not columnNames:  # query on the whole table
                wholeTable = True
                try:
                    columnNames = table_c.columnsOrder  # see ResourceStatusDB for example
                except AttributeError:
                    columns = table_c.__table__.columns  # retrieve the column names
                    columnNames = [str(column).split(".")[1] for column in columns]
                select = Query(table_c, session=session)
            else:  # query only the selected columns
                wholeTable = False
                columns = [getattr(table_c, column) for column in columnNames]
                select = Query(columns, session=session)

            # query conditions
            for columnName, columnValue in params.items():
                if not columnValue:
                    continue
                column_a = getattr(table_c, columnName.lower())
                if isinstance(columnValue, (list, tuple)):
                    select = select.filter(column_a.in_(list(columnValue)))
                elif isinstance(columnValue, six.string_types + (datetime.datetime, bool)):
                    select = select.filter(column_a == columnValue)
                else:
                    self.log.error("type(columnValue) == %s" % type(columnValue))
            if older:
                column_a = getattr(table_c, older[0].lower())
                select = select.filter(column_a < older[1])
            if newer:
                column_a = getattr(table_c, newer[0].lower())
                select = select.filter(column_a > newer[1])
            if order:
                order = [order] if isinstance(order, six.string_types) else list(order)
                column_a = getattr(table_c, order[0].lower())
                if len(order) == 2 and order[1].lower() == "desc":
                    select = select.order_by(desc(column_a))
                else:
                    select = select.order_by(column_a)
            if limit:
                select = select.limit(int(limit))

            # querying
            selectionRes = select.all()

            # handling the results
            if wholeTable:
                selectionResToList = [res.toList() for res in selectionRes]
            else:
                selectionResToList = [[getattr(res, col) for col in columnNames] for res in selectionRes]

            finalResult = S_OK(selectionResToList)

            finalResult["Columns"] = columnNames
            return finalResult

        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("select: unexpected exception", lException=e)
            return S_ERROR("select: unexpected exception %s" % e)
        finally:
            session.close()

    def delete(self, table, params):
        """
        :param table: table from where to delete
        :type table: str
        :param params: dictionary of which line(s) to delete
        :type params: dict

        :return: S_OK() || S_ERROR()
        """
        session = self.sessionMaker_o()

        found = False
        for ext in self.extensions:
            try:
                table_c = getattr(__import__(ext + self.__class__.__module__, globals(), locals(), [table]), table)
                found = True
                break
            except (ImportError, AttributeError):
                continue
        # If not found in extensions, import it from DIRAC base (this same module).
        if not found:
            table_c = getattr(__import__(self.__class__.__module__, globals(), locals(), [table]), table)

        # handling query conditions found in 'Meta'
        older = params.get("Meta", {}).get("older", None)
        newer = params.get("Meta", {}).get("newer", None)
        order = params.get("Meta", {}).get("order", None)
        limit = params.get("Meta", {}).get("limit", None)
        params.pop("Meta", None)

        try:
            deleteQuery = Query(table_c, session=session)
            for columnName, columnValue in params.items():
                if not columnValue:
                    continue
                column_a = getattr(table_c, columnName.lower())
                if isinstance(columnValue, (list, tuple)):
                    deleteQuery = deleteQuery.filter(column_a.in_(list(columnValue)))
                elif isinstance(columnValue, six.string_types + (datetime.datetime, bool)):
                    deleteQuery = deleteQuery.filter(column_a == columnValue)
                else:
                    self.log.error("type(columnValue) == %s" % type(columnValue))
            if older:
                column_a = getattr(table_c, older[0].lower())
                deleteQuery = deleteQuery.filter(column_a < older[1])
            if newer:
                column_a = getattr(table_c, newer[0].lower())
                deleteQuery = deleteQuery.filter(column_a > newer[1])
            if order:
                order = [order] if isinstance(order, six.string_types) else list(order)
                column_a = getattr(table_c, order[0].lower())
                if len(order) == 2 and order[1].lower() == "desc":
                    deleteQuery = deleteQuery.order_by(desc(column_a))
                else:
                    deleteQuery = deleteQuery.order_by(column_a)
            if limit:
                deleteQuery = deleteQuery.limit(int(limit))

            res = deleteQuery.delete(synchronize_session=False)  # FIXME: unsure about it
            session.commit()
            return S_OK(res)

        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("delete: unexpected exception", lException=e)
            return S_ERROR("delete: unexpected exception %s" % e)
        finally:
            session.close()



""" BarGraph represents bar graphs with vertical bars both simple
    and stacked.

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime

from pylab import setp
from matplotlib.patches import Polygon
from matplotlib.dates import date2num

from DIRAC.Core.Utilities.Graphs.PlotBase import PlotBase
from DIRAC.Core.Utilities.Graphs.GraphUtilities import (
    to_timestamp,
    pixelToPoint,
    PrettyDateLocator,
    PrettyDateFormatter,
    PrettyScalarFormatter,
)


__RCSID__ = "$Id$"


class BarGraph(PlotBase):

    """
    The BarGraph class is a straightforward bar graph; given a dictionary
    of values, it takes the keys as the independent variable and the values
    as the dependent variable.
    """

    def __init__(self, data, ax, prefs, *args, **kw):

        PlotBase.__init__(self, data, ax, prefs, *args, **kw)
        if "span" in self.prefs:
            self.width = self.prefs["span"]
        else:
            self.width = 1.0
            if self.gdata.key_type == "time":
                # Try to guess the time bin span
                nKeys = self.gdata.getNumberOfKeys()
                self.width = (max(self.gdata.all_keys) - min(self.gdata.all_keys)) / (nKeys - 1)

    def draw(self):

        PlotBase.draw(self)
        self.x_formatter_cb(self.ax)

        if self.gdata.isEmpty():
            return None

        tmp_x = []
        tmp_y = []

        # Evaluate the bar width
        width = float(self.width)
        if self.gdata.key_type == "time":
            # width = (1 - self.bar_graph_space) * width / 86400.0
            width = width / 86400.0
            offset = 0
        elif self.gdata.key_type == "string":
            self.bar_graph_space = 0.1
            width = (1 - self.bar_graph_space) * width
            offset = self.bar_graph_space / 2.0
        else:
            offset = 0

        start_plot = 0
        end_plot = 0
        if "starttime" in self.prefs and "endtime" in self.prefs:
            start_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["starttime"])))
            end_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["endtime"])))

        nKeys = self.gdata.getNumberOfKeys()
        tmp_b = []
        if "log_yaxis" in self.prefs:
            tmp_b = [0.001] * nKeys
            ymin = 0.001
        else:
            tmp_b = [0.0] * nKeys
            ymin = 0.0

        self.polygons = []
        self.lines = []
        labels = self.gdata.getLabels()
        labels.reverse()

        # If it is a simple plot, no labels are used
        # Evaluate the most appropriate color in this case
        if self.gdata.isSimplePlot():
            labels = [("SimplePlot", 0.0)]
            color = self.prefs.get("plot_color", "Default")
            if color.find("#") != -1:
                self.palette.setColor("SimplePlot", color)
            else:
                labels = [(color, 0.0)]

        seq_b = [(self.gdata.max_num_key + width, 0.0), (self.gdata.min_num_key, 0.0)]
        zorder = 0.0
        dpi = self.prefs.get("dpi", 100)
        for label, num in labels:
            color = self.palette.getColor(label)
            ind = 0
            tmp_x = []
            tmp_y = []
            tmp_t = []
            plot_data = self.gdata.getPlotNumData(label)
            for key, value, error in plot_data:
                if value is None:
                    value = 0.0

                tmp_x.append(offset + key)
                # tmp_y.append(ymin)
                tmp_y.append(0.001)
                tmp_x.append(offset + key)
                tmp_y.append(float(value) + tmp_b[ind])
                tmp_x.append(offset + key + width)
                tmp_y.append(float(value) + tmp_b[ind])
                tmp_x.append(offset + key + width)
                # tmp_y.append(ymin)
                tmp_y.append(0.001)
                tmp_t.append(float(value) + tmp_b[ind])
                ind += 1
            seq_t = list(zip(tmp_x, tmp_y))
            seq = seq_t + seq_b
            poly = Polygon(seq, facecolor=color, fill=True, linewidth=pixelToPoint(0.2, dpi), zorder=zorder)
            self.ax.add_patch(poly)
            self.polygons.append(poly)
            tmp_b = list(tmp_t)
            zorder -= 0.1

        tight_bars_flag = self.prefs.get("tight_bars", False)
        if tight_bars_flag:
            setp(self.polygons, linewidth=0.0)

        # pivots = keys
        # for idx in range(len(pivots)):
        #    self.coords[ pivots[idx] ] = self.bars[idx]

        ymax = max(tmp_b)
        ymax *= 1.1

        if "log_yaxis" in self.prefs:
            ymin = 0.001
        else:
            ymin = min(min(tmp_b), 0.0)
            ymin *= 1.1

        xmax = max(tmp_x)
        if self.log_xaxis:
            xmin = 0.001
        else:
            xmin = 0

        ymin = self.prefs.get("ymin", ymin)
        ymax = self.prefs.get("ymax", ymax)
        xmin = self.prefs.get("xmin", xmin)
        xmax = self.prefs.get("xmax", xmax)

        self.ax.set_xlim(xmin=xmin, xmax=xmax + offset)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        if self.gdata.key_type == "time":
            if start_plot and end_plot:
                self.ax.set_xlim(xmin=start_plot, xmax=end_plot)
            else:
                self.ax.set_xlim(xmin=min(tmp_x), xmax=max(tmp_x))

    def x_formatter_cb(self, ax):
        if self.gdata.key_type == "string":
            smap = self.gdata.getStringMap()
            reverse_smap = {}
            for key, val in smap.items():
                reverse_smap[val] = key
            ticks = sorted(smap.values())
            ax.set_xticks([i + 0.5 for i in ticks])
            ax.set_xticklabels([reverse_smap[i] for i in ticks])
            labels = ax.get_xticklabels()
            ax.grid(False)
            if self.log_xaxis:
                xmin = 0.001
            else:
                xmin = 0
            ax.set_xlim(xmin=xmin, xmax=len(ticks))
        elif self.gdata.key_type == "time":

            # ax.set_xlim( xmin=self.begin_num,xmax=self.end_num )
            dl = PrettyDateLocator()
            df = PrettyDateFormatter(dl)
            ax.xaxis.set_major_locator(dl)
            ax.xaxis.set_major_formatter(df)
            ax.xaxis.set_clip_on(False)
            sf = PrettyScalarFormatter()
            ax.yaxis.set_major_formatter(sf)
            # labels = ax.get_xticklabels()

        else:
            return None



""" An utility to load modules and objects in DIRAC and extensions, being sure that the extensions are considered
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import collections
import os
import re

import pkgutil
import six

import DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities import DIRACSingleton
from DIRAC.Core.Utilities.Extensions import extensionsByPriority, recurseImport


@six.add_metaclass(DIRACSingleton.DIRACSingleton)
class ObjectLoader(object):
    """Class for loading objects. Example:

    from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
    ol = ObjectLoader()
    ol.loadObject('TransformationSystem.Client.TransformationClient')
    """

    def __init__(self, baseModules=False):
        """init"""
        # We save the original arguments in case
        # we need to reinitialize the rootModules
        # CAUTION: we cant do it after doing
        # baseModules = ['DIRAC']
        # because then baseModules, self.baseModules, and __rootModules
        # are the same and edited in place by __generateRootModules !!
        # (Think of it, it's a binding to a list)
        self.originalBaseModules = baseModules
        self._init(baseModules)

    def _init(self, baseModules):
        """Actually performs the initialization"""
        if not baseModules:
            baseModules = ["DIRAC"]
        self.__rootModules = baseModules
        self.__objs = {}
        self.__generateRootModules(baseModules)

    def reloadRootModules(self):
        """Retrigger the initialization of the rootModules.

        This should be used with care.
        Currently, its only use is (and should stay) to retrigger
        the initialization after the CS has been fully initialized in
        LocalConfiguration.enableCS
        """
        # Load the original baseModule argument that was given
        # to the constructor
        baseModules = self.originalBaseModules
        # and replay the init sequence
        self._init(baseModules)

    def __rootImport(self, modName, hideExceptions=False):
        """Auto search which root module has to be used"""
        for rootModule in self.__rootModules:
            impName = modName
            if rootModule:
                impName = "%s.%s" % (rootModule, impName)
            gLogger.debug("Trying to load %s" % impName)
            result = recurseImport(impName, hideExceptions=hideExceptions)
            if not result["OK"]:
                return result
            if result["Value"]:
                return S_OK((impName, result["Value"]))
        return S_OK()

    def __generateRootModules(self, baseModules):
        """Iterate over all the possible root modules"""
        self.__rootModules = baseModules
        for rootModule in reversed(extensionsByPriority()):
            if rootModule not in self.__rootModules:
                self.__rootModules.append(rootModule)
        self.__rootModules.append("")

        # Reversing the order because we want first to look in the extension(s)
        self.__rootModules.reverse()

    def loadModule(self, importString, hideExceptions=False):
        """Load a module from an import string"""
        result = self.__rootImport(importString, hideExceptions=hideExceptions)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR(DErrno.EIMPERR, "No module %s found" % importString)
        return S_OK(result["Value"][1])

    def loadObject(self, importString, objName=False, hideExceptions=False):
        """Load an object from inside a module"""
        if not objName:
            objName = importString.split(".")[-1]

        result = self.loadModule(importString, hideExceptions=hideExceptions)
        if not result["OK"]:
            return result
        modObj = result["Value"]
        try:
            result = S_OK(getattr(modObj, objName))
            result["ModuleFile"] = modObj.__file__
            return result
        except AttributeError:
            return S_ERROR(DErrno.EIMPERR, "%s does not contain a %s object" % (importString, objName))

    def getObjects(self, modulePath, reFilter=None, parentClass=None, recurse=False, continueOnError=False):
        """Search for modules under a certain path

        modulePath is the import string needed to access the parent module.
        Root modules will be included automatically (like DIRAC). For instance "ConfigurationSystem.Service"

        reFilter is a regular expression to filter what to load. For instance ".*Handler"
        parentClass is a class object from which the loaded modules have to import from. For instance RequestHandler

        :param continueOnError: if True, continue loading further module even if one fails
        """
        modules = collections.OrderedDict()
        if isinstance(reFilter, six.string_types):
            reFilter = re.compile(reFilter)

        for rootModule in self.__rootModules:
            impPath = modulePath
            if rootModule:
                impPath = "%s.%s" % (rootModule, impPath)
            gLogger.debug("Trying to load %s" % impPath)

            result = recurseImport(impPath)
            if not result["OK"]:
                return result
            if not result["Value"]:
                continue
            parentModule = result["Value"]
            gLogger.verbose("Loaded module %s at %s" % (impPath, parentModule.__path__))

            for _modLoader, modName, isPkg in pkgutil.walk_packages(parentModule.__path__):
                if reFilter and not reFilter.match(modName):
                    continue
                if isPkg:
                    if recurse:
                        result = self.getObjects(
                            "%s.%s" % (modulePath, modName), reFilter=reFilter, parentClass=parentClass, recurse=recurse
                        )
                        if not result["OK"]:
                            return result
                        modules.update(result["Value"])
                    continue
                modKeyName = "%s.%s" % (modulePath, modName)
                if modKeyName in modules:
                    continue
                fullName = "%s.%s" % (impPath, modName)
                result = recurseImport(fullName)
                if not result["OK"]:
                    if continueOnError:
                        gLogger.error(
                            "Error loading module but continueOnError is true",
                            "module %s error %s" % (fullName, result),
                        )
                        continue
                    return result
                if not result["Value"]:
                    continue

                modClass = getattr(result["Value"], modName, None)
                if not modClass:
                    gLogger.warn("%s does not contain a %s object" % (fullName, modName))
                    continue

                if parentClass and not issubclass(modClass, parentClass):
                    continue

                modules[modKeyName] = modClass

        return S_OK(modules)


def loadObjects(path, reFilter=None, parentClass=None):
    """
    :param str path: the path to the syetem for example: DIRAC/AccountingSystem
    :param object reFilter: regular expression used to found the class
    :param object parentClass: class instance
    :return: dictionary containing the name of the class and its instance
    """
    if not reFilter:
        reFilter = re.compile(r".*[a-z1-9]\.py$")
    pathList = List.fromChar(path, "/")

    objectsToLoad = {}
    # Find which object files match
    for parentModule in extensionsByPriority():
        if six.PY3:
            objDir = os.path.join(os.path.dirname(os.path.dirname(DIRAC.__file__)), parentModule, *pathList)
        else:
            objDir = os.path.join(DIRAC.rootPath, parentModule, *pathList)
        if not os.path.isdir(objDir):
            continue
        for objFile in os.listdir(objDir):
            if reFilter.match(objFile):
                pythonClassName = objFile[:-3]
                if pythonClassName not in objectsToLoad:
                    gLogger.info("Adding to load queue %s/%s/%s" % (parentModule, path, pythonClassName))
                    objectsToLoad[pythonClassName] = parentModule

    # Load them!
    loadedObjects = {}

    for pythonClassName in objectsToLoad:
        parentModule = objectsToLoad[pythonClassName]
        try:
            # Where parentModule can be DIRAC, pathList is something like [ "AccountingSystem", "Client", "Types" ]
            # And the python class name is.. well, the python class name
            objPythonPath = "%s.%s.%s" % (parentModule, ".".join(pathList), pythonClassName)
            objModule = __import__(objPythonPath, globals(), locals(), pythonClassName)
            objClass = getattr(objModule, pythonClassName)
        except Exception as e:
            gLogger.error("Can't load type", "%s/%s: %s" % (parentModule, pythonClassName, str(e)))
            continue
        if parentClass == objClass:
            continue
        if parentClass and not issubclass(objClass, parentClass):
            gLogger.warn("%s is not a subclass of %s. Skipping" % (objClass, parentClass))
            continue
        gLogger.info("Loaded %s" % objPythonPath)
        loadedObjects[pythonClassName] = objClass

    return loadedObjects



"""
  Add-on to make the MonitoringClient works with an IOLoop from a Tornado Server
"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import tornado.ioloop
from DIRAC import gLogger


class MonitoringFlusherTornado(object):
    """
    This class flushes all monitoring clients registered
    Works with the Tornado IOLoop
    """

    def __init__(self):
        self.__mcList = []
        gLogger.info("Using MonitoringClient in IOLoop mode")
        # Here we don't need to use IOLoop.current(), tornado will attach periodic callback to the current IOLoop himself
        # We set callback every 5 minnutes
        tornado.ioloop.PeriodicCallback(self.flush, 300000).start()

    def flush(self, allData=False):
        gLogger.info("Flushing monitoring")
        for mc in self.__mcList:
            mc.flush(allData)

    def registerMonitoringClient(self, mc):
        if mc not in self.__mcList:
            self.__mcList.append(mc)
